from deluca.lung.core import Controller, ControllerState
from deluca.lung.utils import BreathWaveform
from deluca.lung.core import DEFAULT_DT
from deluca.lung.core import proper_time
import deluca.core
import jax
import jax.numpy as jnp
import flax.linen as fnn

class PIDControllerState(deluca.Obj):
    P: float = 0.
    I: float = 0.    
    D: float = 0.
    time: float = float("inf")
    steps: int = 0
    dt: float = DEFAULT_DT

class PID_network(fnn.Module):
    @fnn.compact
    def __call__(self, x):
        x = fnn.Dense(features=1, use_bias=False, name=f"K")(x)
        return x

# generic PID controller
class PID(Controller):
    model: fnn.module = deluca.field(PID_network, jaxed=False)
    params: jnp.array = deluca.field(jaxed=True) # jnp.array([3.0, 4.0, 0.0]
    waveform: deluca.Obj = deluca.field(BreathWaveform.create(), jaxed=False)
    RC: float = deluca.field(0.5, jaxed=False)
    dt: float = deluca.field(0.03, jaxed=False)

    def setup(self):
        self.model = PID_network()
        if self.params is None:
            self.params = self.model.init(
                jax.random.PRNGKey(0), jnp.ones([3,])
            )["params"]   

    def init(self):
        state = PIDControllerState()
        return state

    def __call__(self, controller_state, obs):
        pressure, t = obs.predicted_pressure, obs.time
        target = self.waveform.at(t)
        err = jnp.array(target - pressure)

        decay = jnp.array(self.dt / (self.dt + self.RC))

        P, I, D = controller_state.P, controller_state.I, controller_state.D
        next_P = err
        next_I = I + decay * (err - I)
        next_D = D + decay * (err - P - D)
        controller_state = controller_state.replace(P=next_P, I=next_I, D=next_D)

        next_coef = jnp.array([next_P, next_I, next_D])
        u_in = self.model.apply({"params": self.params}, next_coef)
        u_in = jax.lax.clamp(0.0, u_in.astype(jnp.float32), 100.0)

        # update controller_state
        new_dt = jnp.max(jnp.array([DEFAULT_DT, t - proper_time(controller_state.time)]))
        new_time = t
        new_steps = controller_state.steps + 1
        controller_state = controller_state.replace(time=new_time, steps=new_steps, dt=new_dt)

        return controller_state, u_in



import torch

from deluca.lung import Controller


class IBonusCubic(Controller):
    '''
    Trainable I controller with cubic bonus function of time for jaxed parameters A,B
    '''

    def __init__(self, K_I=2.0, K_P = 1.0, A=-1.0, B= 0.0, C=0.0, D = 0.0,  RC=0.5, **kwargs):
        self.init_K_I = self.tensor(K_I)
        self.init_K_P = self.tensor(K_P)
        self.init_A = self.tensor(A)
        self.init_B = self.tensor(B)
        self.init_C = self.tensor(C)
        self.init_D = self.tensor(D)

        self.K_I = torch.nn.Parameter(self.init_K_I)
        self.K_P = torch.nn.Parameter(self.init_K_P)
        self.A = torch.nn.Parameter(self.init_A)
        self.B = torch.nn.Parameter(self.init_B)
        self.C = torch.nn.Parameter(self.init_C)
        self.D = torch.nn.Parameter(self.init_D)
        self.RC = RC

        self.reset()

    def reset(self):
        self.I = self.tensor(0.0)


    def compute_action(self, state, t):
        err = torch.tensor(self.waveform.at(t)) - state
        dt = self.dt(t)
        decay = dt / (dt + self.RC)

        t_cycle = self.cycle_phase(t) + 1

        self.I = self.I + self.tensor(decay * (err - self.I))


        poly = self.A * t_cycle **3 + self.B *t_cycle**2 + self.C * t_cycle + self.D

        u_in = torch.clamp(self.K_I * self.I + self.K_P * err + poly, min=0.0, max=100.0)

        return (u_in, self.u_out(t))

# generic PID controller
class LookaheadPID(Controller):
    def __init__(self, K=[3, 4, 0], RC=0.5, pressure_history = 4, lookahead = 1,**kwargs):
        self.init_K = self.tensor(K)

        self.K = torch.nn.Parameter(self.init_K)
        self.RC = RC
        self.pressure_history, self.lookahead = pressure_history, lookahead

        self.reset()

    def reset(self):
        self.coef = self.tensor([0.0, 0.0])
        self.last_states = [torch.tensor(0.) for _ in range(self.pressure_history)]

    @property
    def P(self):
        return self.coef[0]

    @property
    def I(self):
        return self.coef[1]

    def compute_action(self, state, t):
        self.last_states.pop(0)
        self.last_states.append(state)
        delta_state = sum(list(map(lambda x: x[0]-x[1], zip(self.last_states[1:], self.last_states[:-1]))))/(self.pressure_history-1)
        err = torch.tensor(self.waveform.at(t)) - state
        dt = self.dt(t)
        decay = dt / (dt + self.RC)

        self.coef = self.tensor(
            [
                err,
                (1 - decay) * self.I + decay * err,
                sum([
                    self.waveform.at(t + dt * i) - (state + delta_state * i) for i in range(self.lookahead)
                ]) / self.lookahead
            ]
        )

        u_in = torch.clamp(torch.dot(self.K, self.coef), min=0.0, max=100.0)

        return (u_in, self.u_out(t))


class IBonus(Controller):
    '''
    Trainable I controller with bonus of the form A t^{-B} for jaxed parameters A,B
    '''

    def __init__(self, K_I=2.0, A=1.0, B= 1.0, RC=0.5, **kwargs):
        self.init_K_I = self.tensor(K_I)
        self.init_A = self.tensor(A)
        self.init_B = self.tensor(B)

        self.K_I = torch.nn.Parameter(self.init_K_I)
        self.A = torch.nn.Parameter(self.init_A)
        self.B = torch.nn.Parameter(self.init_B)
        self.RC = RC

        self.reset()

    def reset(self):
        self.I = self.tensor(0.0)


    def compute_action(self, state, t):
        err = torch.tensor(self.waveform.at(t)) - state
        dt = self.dt(t)
        decay = dt / (dt + self.RC)

        t_cycle = self.cycle_phase(t) + 1

        self.I = self.I + self.tensor(decay * (err - self.I))

        u_in = torch.clamp(self.K_I * self.I + self.A * torch.pow(t_cycle, -self.B), min=0.0, max=100.0)

        return (u_in, self.u_out(t))



import numpy as np
import jax
import jax.numpy as jnp
import deluca
from deluca.lung.core import Controller
from deluca.lung.utils.core import BreathWaveform
from deluca.lung.utils.core import Phase
from deluca.lung.core import DEFAULT_DT
from deluca.lung.core import proper_time

class Expiratory(Controller):
    waveform: BreathWaveform = deluca.field(BreathWaveform, jaxed=False)
    def __call__(self, state, obs, *args, **kwargs):
        pressure, time = obs.predicted_pressure, obs.time
        phase = self.waveform.phase(time)
        u_out = jnp.zeros_like(phase)
        u_out = jax.lax.cond(jax.numpy.logical_or(jnp.equal(phase, Phase.RAMP_DOWN.value), 
                                                  jnp.equal(phase, Phase.PEEP.value)),
                             lambda x : 1,
                             lambda x: x,
                             u_out)

        new_dt = jnp.max(jnp.array([DEFAULT_DT, time - proper_time(state.time)]))
        new_time = time
        new_steps = state.steps + 1
        state = state.replace(time=new_time, steps=new_steps, dt=new_dt)
        return state, u_out



import os
import pickle
import matplotlib.pyplot as plt

from deluca.lung.utils import Analyzer


def save_data_and_plot(result, directory, name):
    if not os.path.exists(directory):
        os.makedirs(directory)
    pickle.dump(result, open(f"{directory}/{name}.pkl", "wb"))
    Analyzer(result).plot()
    plt.savefig(f"{directory}/{name}.png")


import torch

from deluca.lung.core import TorchLungEnv


class CNN(TorchLungEnv):
    def __init__(
        self,
        featurizer,
        out_channels=10,
        kernel=(1, 2),
        stride=1,
        pool_kernel=(1, 2),
        pool_stride=2,
    ):
        self.featurizer = featurizer
        self.in_dim = featurizer.u_window + featurizer.p_window

        out_dim_conv = int((self.in_dim - kernel[1]) / stride + 1)
        out_dim_pool = int((out_dim_conv - pool_kernel[1]) / pool_stride + 1)
        self.layer1 = torch.nn.Sequential(
            torch.nn.Conv2d(1, out_channels, kernel_size=kernel, stride=stride)
        )
        self.fc1 = torch.nn.Linear(out_dim_conv * out_channels, 1)

    def reset(self):
        self.featurizer.reset()
        self.pressure = 0

    def train_step(self, x):
        if len(x.shape) == 1:
            x = torch.reshape(x, (1, x.shape[0]))
        x = torch.reshape(x, (x.shape[0], 1, 1, x.shape[1]))
        out = self.layer1(x)
        out = out.reshape(out.size(0), -1)
        out = self.fc1(out)
        return out

    def forward(self, x):
        return self.train_step(x)

    def step(self, u_in, u_out):
        features = self.featurizer(u_in, u_out, self.pressure)
        self.pressure = self.featurizer.p_scaler.inverse_transform(
            self.train_step(features)
        ).squeeze()
        return self.pressure




import torch
import numpy as np
from deluca.lung.core import Controller
from deluca.lung.utils.core import BreathWaveform
from deluca.lung.utils.core import Phase


class ResidualExplorer(Controller):
    def __init__(self, base_controller, waveform, **kwargs):
        self.base_controller = base_controller

        self.delta_range = (-30, 30)
        self.duration_range = (0.1, 0.5)

        self.boundary_prob = 0.15
        self.boundary_delta_range = (0, 100)
        self.boundary_duration_range = (0.05, 0.3)

        self.needs_refresh = False

        self.waveform = waveform
        self.refresh_exploration_perturbation()

    def refresh_exploration_perturbation(self):
        if np.random.random() < self.boundary_prob:
            self.resample_boundary_delta()
        else:
            self.resample_delta()

    def resample_delta(self):
        # random triangular bump
        height = np.random.uniform(*self.delta_range)
        duration = np.random.uniform(*self.duration_range)

        self.set_delta(height, duration)

    def set_delta(self, height, duration):
        t_min, t_max = 0, self.waveform._keypoints[2]

        t_begin = np.random.uniform(t_min, t_max - duration)
        t_mid = t_begin + 0.5 * duration
        t_end = t_begin + duration

        self.xp = np.array([t_min, t_begin, t_mid, t_end, t_max])
        self.fp = np.array([0.0, 0.0, height, 0.0, 0.0])

    def resample_boundary_delta(self):
        # random lean-to triangular bump at inspiratory boundary
        height = np.random.uniform(*self.boundary_delta_range)
        duration = np.random.uniform(*self.boundary_duration_range)

        self.set_boundary_delta(height, duration)

    def set_boundary_delta(self, height, duration):
        t_min, t_max = 0, self.waveform._keypoints[2]

        self.xp = np.array([t_min, duration, t_max])
        self.fp = np.array([height, 0, 0])

    def action(self, pressure, target, t):
        base_u_in = self.base_controller(pressure, target, t)

        if isinstance(base_u_in, torch.Tensor):
            base_u_in = base_u_in.detach().numpy()

        if (
            self.waveform.phase(t) == Phase.PEEP or self.waveform.phase(t) == Phase.RAMP_DOWN
        ):  # if base exhales, just follow base
            self.u_in = base_u_in
            if self.needs_refresh:
                self.refresh_exploration_perturbation()
                self.needs_refresh = False
        else:  # if base inhales, add residual to cycle phase
            self.needs_refresh = True
            self.u_in = base_u_in + np.interp(t % self.waveform.period, self.xp, self.fp)

        return np.clip(self.u_in, self.min, self.max)



from deluca.lung.core import Controller
from deluca.lung.controllers._expiratory import Expiratory
from deluca.lung.environments._physical_lung import PhysicalLung
from deluca.lung.environments._balloon_lung import BalloonLung
from deluca.lung.utils.core import BreathWaveform
from deluca.lung.utils.scripts.save_data_and_plot import save_data_and_plot

import os
import click
import dill as pickle
import datetime
import time
import tqdm
import jax
import jax.numpy as jnp

def run_controller(
    controller,
    R,
    C,
    T=1000,
    dt=0.03,
    abort=60,
    env=None,
    waveform=None,
    use_tqdm=False,
    directory=None,
):
    env = env or BalloonLung()
    waveform = waveform or BreathWaveform.create()
    expiratory = Expiratory.create(waveform=waveform)

    result = locals()

    controller_state = controller.init()
    expiratory_state = expiratory.init()

    tt = range(T)
    if use_tqdm:
        tt = tqdm.tqdm(tt, leave=False)

    timestamps = jnp.zeros(T)
    pressures = jnp.zeros(T)
    flows = jnp.zeros(T)
    u_ins = jnp.zeros(T)
    u_outs = jnp.zeros(T)

    state, obs = env.reset()

    try:
        for i, _ in enumerate(tt):
            pressure = obs.predicted_pressure
            if env.should_abort():
                break

            controller_state, u_in = controller.__call__(controller_state, obs)
            expiratory_state, u_out = expiratory.__call__(expiratory_state, obs)
            state, obs = env(state, (u_in, u_out))

            timestamps = jax.ops.index_update(timestamps, i, env.time(state) - dt)
            u_ins = jax.ops.index_update(u_ins, i, u_in)
            u_outs = jax.ops.index_update(u_outs, i, u_out)
            pressures = jax.ops.index_update(pressures, i, pressure)
            flows = jax.ops.index_update(flows, i, env.flow)

            env.wait(max(dt - env.dt, 0))

    finally:
        env.cleanup()

    timeseries = {
        "timestamp": jnp.array(timestamps),
        "pressure": jnp.array(pressures),
        "flow": jnp.array(flows),
        "target": waveform.at(timestamps),
        "u_in": jnp.array(u_ins),
        "u_out": jnp.array(u_outs),
    }

    for key, val in timeseries.items():
        timeseries[key] = val[: T + 1]

    result["timeseries"] = timeseries

    if directory is not None:
        if not os.path.exists(directory):
            os.makedirs(directory)
        timestamp = datetime.datetime.now().strftime("%Y-%m-%dT%H-%M-%S")
        pickle.dump(result, open(f"{directory}/{timestamp}.pkl", "wb"))

    return result

def run_controller_scan(
    controller,
    R,
    C,
    T=1000,
    dt=0.03,
    abort=60,
    env=None,
    waveform=None,
    use_tqdm=False,
    directory=None,
):
    env = env or BalloonLung()
    waveform = waveform or BreathWaveform.create()
    expiratory = Expiratory.create(waveform=waveform)

    result = locals()

    controller_state = controller.init()
    expiratory_state = expiratory.init()

    tt = range(T)
    if use_tqdm:
        tt = tqdm.tqdm(tt, leave=False)

    timestamps = jnp.zeros(T)
    pressures = jnp.zeros(T)
    flows = jnp.zeros(T)
    u_ins = jnp.zeros(T)
    u_outs = jnp.zeros(T)

    state, obs = env.reset()
    xp = jnp.array(waveform.xp)
    fp = jnp.array(waveform.fp)
    period = waveform.period
    dtype = waveform.dtype

    try:
        def loop_over_tt(envState_obs_ctrlState_ExpState_i, dummy_data):
            state, obs, controller_state, expiratory_state, i = envState_obs_ctrlState_ExpState_i
            pressure = obs.predicted_pressure
            # if env.should_abort(): # TODO: how to handle break in scan
            #     break

            controller_state, u_in = controller.__call__(controller_state, obs)
            expiratory_state, u_out = expiratory.__call__(expiratory_state, obs)

            state, obs = env(state, (u_in, u_out))

            timestamps_i = env.time(state) - dt
            u_ins_i = u_in
            u_outs_i = u_out
            pressures_i = pressure
            flows_i = env.flow

            env.wait(max(dt - env.dt, 0))
            return (state, obs, controller_state, expiratory_state, i + 1), (timestamps_i, u_ins_i, u_outs_i, pressures_i, flows_i)

        _, (timestamps, u_ins, u_outs, pressures, flows) = jax.lax.scan(loop_over_tt, (state, obs, controller_state, expiratory_state, 0), jnp.arange(T))
        
    finally:
        env.cleanup()

    timeseries = {
        "timestamp": jnp.array(timestamps),
        "pressure": jnp.array(pressures),
        "flow": jnp.array(flows),
        "target": waveform.at(timestamps),
        "u_in": jnp.array(u_ins),
        "u_out": jnp.array(u_outs),
    }

    for key, val in timeseries.items():
        timeseries[key] = val[: T + 1]

    result["timeseries"] = timeseries

    if directory is not None:
        if not os.path.exists(directory):
            os.makedirs(directory)
        timestamp = datetime.datetime.now().strftime("%Y-%m-%dT%H-%M-%S")
        pickle.dump(result, open(f"{directory}/{timestamp}.pkl", "wb"))

    return result


def run_controller_command_line(path, r, c, peep, pip, bpm, t, dt, abort, name):
    directory = os.path.dirname(os.path.realpath(path))
    controller = Controller.load(path)
    waveform = BreathWaveform(range=(peep, pip), bpm=bpm)
    pickle.dump(locals(), open(os.path.join(directory, "meta.pkl"), "wb"))

    result = run_controller(
        controller,
        R=r,
        C=c,
        T=t,
        dt=dt,
        abort=abort,
        env=PhysicalLung(),
        waveform=waveform,
        directory=directory,
    )

    save_data_and_plot(result, directory, name)


@click.command()
@click.argument("path", type=click.Path(exists=True), required=1)
@click.option("-R", type=int, default=50, help="R value for phsyical lung")
@click.option("-C", type=int, default=10, help="C value for phsyical lung")
@click.option("--PEEP", type=int, default=5, help="PEEP")
@click.option("--PIP", type=int, default=35, help="PIP")
@click.option("--bpm", type=int, default=20, help="bpm")
@click.option("-T", type=int, default=1000, help="Default timesteps")
@click.option("--dt", type=float, default=0.03, help="Time to wait")
@click.option("--abort", type=float, default=60, help="Abort pressure")
@click.option("--name", type=str, default="run", help="Name of experiment")
def _run_controller_command_line(path, r, c, peep, pip, bpm, t, dt, abort, name):
    return run_controller_command_line(path, r, c, peep, pip, bpm, t, dt, abort, name)


# For running on the physical lung from command line with controller path
if __name__ == "__main__":
    _run_controller_command_line()



from deluca.lung.environments._balloon_lung import BalloonLung
from deluca.lung.environments._delay_lung import DelayLung
from deluca.lung.environments._physical_lung import PhysicalLung
from deluca.lung.environments._stitched_sim import StitchedSim
# from deluca.lung.environments._linear import Linear
# from deluca.lung.environments._cnn import CNN
# from deluca.lung.environments._lstm import LSTM
# from deluca.lung.environments._mlp import MLP
# from deluca.lung.environments._single_comp_lung_pid import SingleCompLung
# from deluca.lung.environments._double_comp_lung_pid import DoubleCompLung

__all__ = [
    "BalloonLung",
    "DelayLung",
    "PhysicalLung",
    "StitchedSim",
    # "Linear",
    # "CNN",
    # "MLP",
    # "SingleCompLung",
    # "DoubleCompLung",
]



import os
import pickle
from google.cloud import storage
import hypertune

from deluca.lung.core import ROOT


def dump(obj, args, name, metric, step, bucket_name="vizier_cloud_codelab"):
    directory = "/tmp/results"

    args = vars(args)
    filename = "-".join([f"{k}_{v}" for k, v in args.items()])
    filename = f"{filename}.pkl"

    if not os.path.exists(directory):
        os.makedirs(directory)

    obj = {
        "result": obj,
        "args": args
    }

    path = os.path.join(directory, filename)
    with open(path, "wb") as f:
        pickle.dump(obj, f)

    client = storage.Client()
    bucket = client.get_bucket(bucket_name)
    with open(os.path.join(ROOT, "vizier/jobs", name, "run"), "r") as f:
        run = int(f.read().strip())
    print(run)
    blob = bucket.blob(f"{name}/{run}/{filename}")
    blob.upload_from_filename(filename=path)

    hpt = hypertune.HyperTune()
    hpt.report_hyperparameter_tuning_metric(
        hyperparameter_metric_tag="score",
        metric_value=metric,
        global_step=step,
    )



import numpy as np
import torch
from deluca.lung.controllers.core import Controller
from deluca.lung.controllers import PID
import time

class MPC(Controller):
    def __init__(self, sim, MPC_lookahead=10, output_length=1, optimization_params=None, verbose=False,
                 initializer='ZEROS', init_params=None, **kwargs):
        super().__init__()

        self.sim = sim
        self.MPC_lookahead = MPC_lookahead
        self.output_length = output_length
        self.optimization_params = optimization_params
        self.verbose = verbose
        self.initializer = initializer
        self.dthard = 0.03
        self.init_params = init_params
        self.reset()

    def reset(self):
        self.plan = []
        self.immediate_counter = 0
        self.counter = 0

    def compute_action(self, state, t):
        #print(t)
        a = time.time()
        decay = self.waveform.decay(t)
        self.sim.set_current_pressure(state)
        # returns a copy of the state as it will be needed to reset sim
        if decay is None:
            if self.counter >= self.immediate_counter:
                #a = time.time()
                self.update_plan(t)
                #print(time.time()-a)
                self.immediate_counter = self.counter+self.output_length
            u_in = self.plan[self.counter]
            # is this necessary?
            u_in = torch.clamp(u_in, min=0., max=100.)
        else:
            self.immediate_counter = self.counter + 1
            u_in = decay
            if self.counter < len(self.plan):
                self.plan[self.counter] = torch.tensor(u_in)
            else:
                self.plan.append(torch.tensor(u_in))
        u_out = self.u_out(t)
        # Update the sim according to new syntax
        self.sim(u_in, u_out, t)
        self.counter += 1
        # print(time.time()-a)
        return u_in, u_out

    def initialization(self, sub_plan, extra_plan_length, trange):
        #print("In initialization")
        #print(sub_plan)
        #print(extra_plan_length)
        #print(trange)
        if self.initializer == 'ZEROS':
            return [torch.tensor(0.0, requires_grad=True) for _ in range(extra_plan_length)]
        elif self.initializer == 'PID':
            PIDvals = self.init_params['PID']
            PIDcont = PID(K=PIDvals, waveform=self.waveform)
            self.sim.cache_state()
            #print(self.sim.state['t_in'])
            #print(self.sim.state['u_history'])
            #print(self.sim.state['p_history'])
            # rollout the subplan
            for i in range(len(sub_plan)):
                self.sim(sub_plan[i], self.u_out(trange[i]), trange[i])
            # now use the controller
            extra_plan = []
            for i in range(len(sub_plan), len(trange)):
                #print(self.sim.pressure.item(), trange[i])
                u_in, u_out = PIDcont(self.sim.pressure.item(), trange[i])
                #print(u_in, u_out)
                extra_plan.append(torch.tensor(u_in.item(), requires_grad=True))
                self.sim(u_in, self.u_out(trange[i]), trange[i])
            #print("aaa", self.sim.state['t_in'])
            #print("aaa", self.sim.state['u_history'])
            #print("aaa", self.sim.state['p_history'])
            self.sim.rewind_state()
            #print(extra_plan)
            return extra_plan
        else:
            print("Initializer not implemented")
            exit()

    def update_plan(self, start_t):
        if self.verbose:
            print("Updating plan")
        # This code finds a good lookahead open loop policy
        extra_plan_length = self.counter + self.MPC_lookahead - len(self.plan)
        sub_plan = self.plan[self.counter:]
        trange = [start_t + i*self.dthard for i in range(self.MPC_lookahead)]
        extra_plan = self.initialization(sub_plan, extra_plan_length, trange)
        self.plan.extend(extra_plan)
        if self.optimization_params:
            epochs = self.optimization_params['epochs']
            schedule = self.optimization_params['schedule']
            optimizer = self.optimization_params['optimizer']
        else:
            epochs = 30
            schedule = {
                0: .5,
            }
            optimizer = 'Adam'
        losses = []
        end_t = start_t + self.MPC_lookahead*self.dthard
        tt = np.linspace(start_t, end_t, self.MPC_lookahead)
        local_plan = self.plan[self.counter:]
        self.sim.cache_state()
        for epoch in range(epochs):
            if epoch in schedule:
                if optimizer == 'SGD':
                    optim = torch.optim.SGD(local_plan, lr=schedule[epoch])
                elif optimizer == 'Adam':
                    optim = torch.optim.Adam(local_plan, lr=schedule[epoch])
                else:
                    print('error')
            loss = torch.tensor(0.0, requires_grad=True)
            counter = 0
            for j, t in enumerate(tt):
                if self.waveform.decay(t) is None:
                    pressure = self.sim.pressure
                    #u_in, u_out = torch.clamp(local_plan[j], 0., 100.), self.u_out(t)
                    u_in, u_out = local_plan[j], self.u_out(t)
                    # new syntax
                    self.sim(u_in, u_out, t)
                    decay = self.waveform.decay(t)
                    if decay is None:
                        loss = loss + torch.square(pressure - self.waveform.at(t))
                        counter += 1
            # loss.backward(retain_graph=True)
            loss.backward()
            optim.step()
            optim.zero_grad()
            #print(epoch, loss)
            normalized_loss = (loss / counter).sqrt()
            losses.append(normalized_loss / len(tt))
            self.sim.rewind_state()
        if self.verbose:
            print(f"Normalized loss at time step {start_t:0.2f} - { normalized_loss:0.4f}")



import torch
import numpy as np

from deluca.lung.core import TorchLungEnv


def PropValve(x):  # copied from Controller.__SimulatedPropValve
    y = 3 * x
    flow_new = 1.0 * (torch.tanh(0.03 * (y - 130)) + 1)
    return torch.clamp(flow_new, 0.0, 1.72)


def Solenoid(x):  # copied from Controller.__SimulatedSolenoid
    if x > 0:
        return x / x
    else:
        return x * 0.0


# balloon physics vent ported from Cohen lab's repository
# Sources:
# https://github.com/CohenLabPrinceton/Ventilator-Dev/blob/master/sandbox/HOWTO_RunController.ipynb
# https://github.com/CohenLabPrinceton/Ventilator-Dev/blob/master/vent/controller/control_module.py
class SingleCompLung(TorchLungEnv):
    def __init__(
        self,
        resistance=6.4,
        compliance=0.1,
        min_volume=0.2,
        peep_valve=5.0,
        PC=40.0,
        RP=1.0,
        leak=False,
        **kwargs
    ):
        # dynamics hyperparameters
        self.min_volume = self.tensor(min_volume)
        self.PC = self.tensor(PC)
        self.RP = self.tensor(RP)
        self.P0 = self.tensor(0.0)
        self.leak = leak
        self.peep_valve = self.tensor(peep_valve)
        self.R = self.tensor(resistance)
        self.C = self.tensor(compliance)
        self.flow = self.tensor(0.0, requires_grad=False)

        # reset states
        self.reset()

    def reset(self):
        self.flow = self.tensor(0.0, requires_grad=False)
        self.volume = self.min_volume
        self.p = (self.flow * self.R) + (self.volume / self.C) + self.P0
        self.pressure = self.tensor(self.p, requires_grad=False)

    def compute_aux_states(self):
        # compute all other state vars, which are just functions of volume
        self.p = (self.flow * self.R) + (self.volume / self.C) + self.P0

        self.pressure = self.tensor(self.p, requires_grad=False)

    def step(self, u_in, u_out):
        if not isinstance(u_in, torch.Tensor):
            u_in = self.tensor(u_in)
        dt = self.dt

        # 2-dimensional action per timestep: PIP/PEEP voltages
        self.flow = torch.clamp(PropValve(u_in), 0, 2) * self.RP

        # update by flow rate
        self.volume = self.volume + self.flow * dt

        # compute and record state
        self.compute_aux_states()

        return self.pressure



# https://pypi.org/project/jax-data/

import jax
import jax.numpy as jnp

import numpy as np

def default_collate_fn(samples):
    X = jnp.array([sample[0] for sample in samples])
    Y = jnp.array([sample[1] for sample in samples])
    return X, Y

class Dataset(object):
    def __init__(self):
        pass

    def __len__(self):
        raise NotImplementedError

    def __getitem__(self, idx):
        raise NotImplementedError

class Dataset_from_XY(Dataset):
    def __init__(self, X,Y):
        self.X = X
        self.Y = Y
    def __len__(self):
        return len(self.X)
    def __getitem__(self, idx):
        return (self.X[idx], self.Y[idx])

class Dataloader_jax(object):
    def __init__(self, dataset, batch_size, shuffle=True, collate_fn=default_collate_fn):
        """
        Adapted from Pytorch Dataloader implementation.

        params:
            - dataset : a class with the __len__ and __getitem__ implemented.
            - batch_size : size of each batch.
            - shuffle : whether to shuffle the dataset upon epoch ending.
            - collate_fn : How the samples are collated.

        Note: To shuffle the dataset, call on_epoch_end().
        """
        self.dataset = dataset
        self.batch_size = batch_size
        self.length = len(self.dataset)
        self.shuffle = shuffle
        self.on_epoch_end()
        self.collate_fn = collate_fn


    def __getitem__(self, idx):
        if len(self) <= idx:
            raise IndexError("Index out of range")
        indices = self.indices[idx * self.batch_size: (idx + 1) * self.batch_size]

        samples = []
        for i in indices:
            data = self.dataset[i]
            samples.append(data)
        return self.collate_fn(samples) 

    def __len__(self):
        return int(jnp.floor(len(self.dataset) / self.batch_size))

    def on_epoch_end(self):
        n = len(self.dataset)
        seq = np.arange(0, n)
        if self.shuffle:
            np.random.shuffle(seq)
            self.indices = seq
        else:
            self.indices = seq


import torch

from deluca.lung.core import TorchLungEnv


class MLP(TorchLungEnv):
    def __init__(self, featurizer, hid_1=10, hid_2=20, droprate=0.3):
        self.featurizer = featurizer
        self.in_dim = featurizer.u_window + featurizer.p_window
        self.network = torch.nn.Sequential(
            torch.nn.Linear(self.in_dim, hid_1),
            torch.nn.Dropout(droprate),
            torch.nn.ReLU(),
            torch.nn.Linear(hid_1, hid_2),
            torch.nn.Dropout(droprate),
            torch.nn.ReLU(),
            torch.nn.Linear(hid_2, 1),
        )

        self.reset()

    def reset(self):
        self.featurizer.reset()
        self.pressure = 0

    def forward(self, x):
        return self.train_step(x)

    def step(self, u_in, u_out):
        """
        For inference and returns unscaled pressure
        """
        features = self.featurizer(u_in, u_out, self.pressure)
        self.pressure = self.featurizer.p_scaler.inverse_transform(
            self.train_step(features)
        ).squeeze()
        return self.pressure

    def train_step(self, features):
        """
        For training and expects/returns scaled features
        """
        return self.network(features)



import numpy as np

from deluca.lung.core import LungEnv


# 2nd attempt at vent simulation
# observed pressure comes from delayed control signal, with back-pressure from deluca.lung
class DelayLung(LungEnv):
    def __init__(
        self, R=10, C=6, delay=10, min_volume=1.5, inertia=0.995, control_gain=0.02, **kwargs
    ):
        # dynamics hyperparameters
        self.R = R
        self.C = C
        self.min_volume = min_volume
        self.inertia = inertia
        self.control_gain = control_gain
        self.delay = delay

        # reset states
        self.r0 = (3 * self.min_volume / (4 * np.pi)) ** (1 / 3)
        self.reset()

    def reset(self):
        super().reset()
        self.volume = self.min_volume
        self.pipe_pressure = 0.

        self.controls_in, self.controls_out = [], []
        self.compute_aux_states()

    def compute_aux_states(self):
        # compute all other state vars, which are just functions of volume
        r = (3 * self.volume / (4 * np.pi)) ** (1 / 3)
        self.vent_pressure = self.C * (1 - (self.r0 / r) ** 6) / (self.r0 ** 2 * r)

        if len(self.controls_in) < self.delay:
            self.pipe_impulse = 0
            self.peep = 0
        else:
            self.pipe_impulse = self.control_gain * self.controls_in[-self.delay]
            self.peep = self.controls_out[-self.delay]

        self.pipe_pressure = self.inertia * self.pipe_pressure + self.pipe_impulse
        self.pressure = np.maximum(0., self.pipe_pressure - self.vent_pressure)

        if self.peep:
            self.pipe_pressure *= 0.995

    def step(self, u_in, u_out):
        u_in = np.maximum(0., u_in)

        self.controls_in.append(u_in)
        self.controls_out.append(u_out)

        # 2-dimensional action per timestep
        flow = self.pressure / self.R

        # update by flow rate
        self.volume = np.maximum(self.volume + flow * self.dt, self.min_volume)

        # compute and record state
        self.compute_aux_states()

        return self.pressure


import torch

from deluca.lung.core import TorchLungEnv


class LSTM(TorchLungEnv):
    def __init__(self, featurizer, device, hidden_size=10, num_layers=1):
        self.featurizer = featurizer
        self.in_dim = featurizer.u_window + featurizer.p_window

        self.device = device
        self.num_layers = num_layers
        self.hidden_size = hidden_size

        self.lstm = torch.nn.LSTM(self.in_dim, hidden_size, num_layers)
        self.fc = torch.nn.Linear(hidden_size, 1)

        self.reset()

    def reset(self):
        self.featurizer.reset()
        self.pressure = 0

    def forward(self, x):
        return self.train_step(x)

    def step(self, u_in, u_out):
        """
        For inference and returns unscaled pressure
        """
        features = self.featurizer(u_in, u_out, self.pressure)
        self.pressure = self.featurizer.p_scaler.inverse_transform(
            self.train_step(features)
        ).squeeze()
        return self.pressure

    def train_step(self, features):
        """
        For training and expects/returns scaled features
        """
        if len(features.shape) == 1:
            features = features.unsqueeze(0)
        out, _ = self.lstm(features.unsqueeze(0))
        out = self.fc(out.squeeze(0))
        return out



import tqdm
import jax.numpy as jnp
from deluca.lung.utils.data.analyzer import Analyzer


class Dataset:
    def __init__(self, paths, inspiratory_only=True, clip=(1, -1), breath_length=29):
        self.data = []

        if isinstance(paths, str):
            paths = [paths]

        for path in tqdm.tqdm(paths):
            analyzer = Analyzer(path)
            if inspiratory_only:
                clips = analyzer.infer_inspiratory_phases()
            else:
                clips = analyzer.infer_breaths()

            for start, end in clips[clip[0] : clip[1]]:
                if end - start < breath_length:
                    continue

                u_in = analyzer.u_in[start : start + breath_length]
                pressure = analyzer.pressure[start : start + breath_length]

                self.data.append((jnp.array(u_in), jnp.array(pressure)))

    def __len__(self):
        return len(self.data)

    def __getitem__(self, index):
        return self.data[index]

    def __add__(self, other):
        self.data.extend(other.data)

    def __iter__(self):
        self.n = 0
        return self

    def __next__(self):
        if self.n < len(self):
            item = self[self.n]
            self.n += 1
            return item
        else:
            raise StopIteration



import numpy as np
import torch
from copy import deepcopy

from deluca.lung.environments.core import Environment


class OldStitchedSim(Environment):
    def __init__(self, inspiratory_model, u_window, p_window, u_scaler, p_scaler):
        super().__init__()

        self.inspiratory_model = inspiratory_model

        self.u_window = u_window
        self.p_window = p_window

        self.u_scaler = u_scaler
        self.p_scaler = p_scaler

        # reset states
        self.reset()

    def reset(self):
        self.scaled_peep = self.inspiratory_model.boundary_dict["0"]().item()
        self.state = {'t_in':0, 'u_history':[], 'p_history':[self.scaled_peep]}
        self.sync_unscaled_pressure()

    def sync_unscaled_pressure(self):
        self.pressure = self.p_scaler.inverse_transform([[self.state['p_history'][-1]]])[0, 0]

    def cache_state(self, return_state=False):
        # set up a snapshot for rewind_state
        state_copy = deepcopy(self.state)
        if return_state:
            return state_copy
        else:
            self.cached_state = state_copy

    def rewind_state(self, state=None):
        # load a history state cached by cache_state
        if state:
            self.state = deepcopy(state)
        else:
            self.state = deepcopy(self.cached_state)
        self.sync_unscaled_pressure()

    def step(self, u_in, u_out, t):

        u_in_scaled = self.u_scaler.transform([[u_in]])[0, 0]
        self.state['u_history'].append(u_in_scaled)

        if u_out == 1:
            if self.state['t_in'] > 0: # reset once per u_out=1
                self.state = {'t_in':0, 'u_history':[], 'p_history':[self.scaled_peep]}
                self.sync_unscaled_pressure()
        else:
            self.state['t_in'] += 1
            t_key = str(self.state['t_in'])

            if t_key in self.inspiratory_model.boundary_dict:  # predict from boundary model
                features = np.concatenate([self.state['u_history'], self.state['p_history']])
                features = torch.tensor(features, dtype=torch.float)
                scaled_pressure = self.inspiratory_model.boundary_dict[t_key](features).item()
            else:  # predict from default model
                features = np.concatenate(
                    [self.state['u_history'][-self.u_window:], self.state['p_history'][-self.p_window:]]
                )
                features = torch.tensor(features, dtype=torch.float)
                scaled_pressure = self.inspiratory_model.default_model(features).item()

            self.state['p_history'].append(scaled_pressure)
            self.sync_unscaled_pressure()



import os

import deluca.core
import jax
import jax.numpy as jnp

ROOT = os.path.dirname(os.path.realpath(os.path.join(__file__, "../..")))

DEFAULT_DT = 0.03


class LungEnv(deluca.Env):
    # @property
    def time(self, state):
        return self.dt * state.steps

    @property
    def dt(self):
        return DEFAULT_DT

    @property
    def physical(self):
        return False

    def should_abort(self):
        return False

    def wait(self, duration):
        pass

    def step(self, u_in, u_out):
        pass

    def cleanup(self):
        pass

class ControllerState(deluca.Obj):
    time: float = float("inf")
    steps: int = 0
    dt: float = DEFAULT_DT

def proper_time(t):
    return jax.lax.cond(t == float('inf'), lambda x : 0., lambda x: x, t)

class Controller(deluca.Agent):
    def __new__(cls, *args, **kwargs):
        obj = deluca.Agent.__new__(cls)

        super(cls, obj).init()

        return obj

    def __getattribute__(self, name):
        attr = object.__getattribute__(self, name)
        if name == "init":
            super(self.__class__, self).init()
        return attr

    def init(self):
        return ControllerState()

    @property
    def max(self):
        return 100

    @property
    def min(self):
        return 0

    def action(self, pressure, target, time):
        return 0



from deluca.lung.controllers._bang_bang import BangBang
from deluca.lung.controllers._expiratory import Expiratory
from deluca.lung.controllers._impulse import Impulse
from deluca.lung.controllers._periodic_impulse import PeriodicImpulse
from deluca.lung.controllers._pid import PID
from deluca.lung.controllers._predestined import Predestined
from deluca.lung.controllers._residual_explorer import ResidualExplorer

__all__ = [
    "BangBang",
    "Expiratory",
    "Impulse",
    "PeriodicImpulse",
    "PID",
    "Predestined",
    "ResidualExplorer"
]



import jax.numpy as jnp


class ShiftScaleTransform:
    # vectors is an array of vectors
    def __init__(self, vectors):
        vectors_concat = jnp.concatenate(vectors)
        self.mean = jnp.mean(vectors_concat)
        self.std = jnp.std(vectors_concat)
        print(self.mean, self.std)

    def _transform(self, x, mean, std):
        return (x - mean) / std

    def _inverse_transform(self, x, mean, std):
        return (x * std) + mean

    def __call__(self, vector):
        return self._transform(vector, self.mean, self.std)

    def inverse(self, vector):
        return self._inverse_transform(vector, self.mean, self.std)



from deluca.lung.utils.data.munger import Munger
import math
import torch
from collections import OrderedDict
import numpy as np
import matplotlib.pyplot as plt
import pickle
import flax
import flax.linen as fnn
import jax
import jax.numpy as jnp
from jax import jit, partial
import functools
import optax
import deluca
from deluca.lung.utils.data.alpha_dropout import Alpha_Dropout

class SNN_jax(fnn.Module):
    u_history: int # = 5 # max(u_window, num_boundary_models)
    u_window: int # = 5 # number of features for default model
    p_window: int # = 3 # number of features for default model
    out_dim: int # = 1
    hidden_dim: int # = 100
    n_layers: int # = 4
    dropout_prob: float # = 0.0
    scale: float = 1.0507009873554804934193349852946
    alpha: float= 1.6732632423543772848170429916717

    @fnn.compact
    def __call__(self, x):
        for i in range(self.n_layers - 1):
            x = jnp.hstack([x[self.u_history - self.u_window: self.u_history], x[-self.p_window:]])
            x = fnn.Dense(features=self.hidden_dim, use_bias=False, name=f"fc{i}")(x)
            x = self.scale*fnn.elu(x, alpha=self.alpha)
            x = Alpha_Dropout(rate=self.dropout_prob, deterministic=False)(x)
        x = fnn.Dense(features=self.out_dim, use_bias=True, name=f"fc{i + 1}")(x)
        return x

class ShallowBoundaryModel_jax(fnn.Module):
    u_history: int # = 5
    input_dim: int # = 2
    out_dim: int # = 1
    hidden_dim: int # = 100
    model_num: int
        
    @fnn.compact
    def __call__(self, x):
        x = jnp.hstack([x[self.u_history - self.input_dim: self.u_history], x[-self.input_dim:]])
        x = fnn.Dense(features=self.hidden_dim, use_bias=False, name="shallow_fc1_model"+str(self.model_num))(x)
        x = fnn.tanh(x)
        x = fnn.Dense(features=self.out_dim, use_bias=True, name="shallow_fc2_model"+str(self.model_num))(x)
        return x

class InspiratoryModel_jax(fnn.Module):
    # boundary_mods: dict = deluca.field(default_factory=dict, jaxed=False)
    # default_mod: fnn.module = deluca.field(SNN_jax, jaxed=False)
    u_history: int = 5 # max(u_window, num_boundary_models)
    p_history: int = 5 # max(p_window, num_boundary_models)
    u_window: int = 5 # number of features for default model
    p_window: int = 3 # number of features for default model
    
    default_out_dim: int = 1
    default_hidden_dim: int = 100
    default_n_layers: int = 4
    default_dropout_prob: float = 0.0

    num_boundary_models: int = 5
    boundary_out_dim: int = 1
    boundary_hidden_dim: int = 100
    seed: int = 0

    def setup(self):
        
        self.default_mod = SNN_jax(u_history=self.u_history,
                                   u_window=self.u_window,
                                   p_window=self.p_window,
                                   out_dim=self.default_out_dim,
                                   hidden_dim=self.default_hidden_dim,
                                   n_layers=self.default_n_layers,
                                   dropout_prob=self.default_dropout_prob)

        self.boundary_mod = ShallowBoundaryModel_jax(u_history=self.u_history,
                                                     input_dim=self.u_history,
                                                     out_dim=self.boundary_out_dim,
                                                     hidden_dim=self.boundary_hidden_dim,
                                                     model_num=0)
        
        self.boundary_mods = {str(i): ShallowBoundaryModel_jax(u_history=self.u_history,
                                                                 input_dim=self.u_history,
                                                                 out_dim=self.boundary_out_dim,
                                                                 hidden_dim=self.boundary_hidden_dim,
                                                                 model_num=i)
                              for i in range(self.num_boundary_models)}
        '''
        self.all_models_dict = self.boundary_mods.copy({str(self.num_boundary_models+1): 
            SNN_jax(u_history=self.u_history,
                    u_window=self.u_window,
                    p_window=self.p_window,
                    out_dim=self.default_out_dim,
                    hidden_dim=self.default_hidden_dim,
                    n_layers=self.default_n_layers,
                    dropout_prob=self.default_dropout_prob)})
        self.all_models = [self.all_models_dict[str(i+1)]
                           for i in range(self.num_boundary_models+1)]
        print(type(self.all_models))
        print(self.all_models)'''

    def call_boundary_model(self, t, features):
        return self.boundary_mods[str(t)](features)

    def call_default_model(self, features):
        return self.default_mod(features)

    # couldn't get this methods to work
    def all_models_call(self, t_features):
        t, features = t_features
        # return jax.lax.switch(t, self.all_models, features)
        return self.all_models[t](features)
        # model = next(flax.traverse_util.TraverseItem(t).iterate(self.all_models))
        # return model(features)

    # couldn't get this methods to work
    def boundary_model(self, features):
        model = partial(ShallowBoundaryModel_jax, u_history=self.u_history,
                                                  input_dim=self.u_history, # wasted parameters
                                                  out_dim=self.boundary_out_dim,
                                                  hidden_dim=self.boundary_hidden_dim)
        return model(features)
        # x = jnp.hstack([x[self.u_history - t: self.u_history], x[-t:]])
        # x = jnp.hstack([jax.lax.dynamic_slice(x, (self.u_history-t,), (t,)),
        #                 jax.lax.dynamic_slice(x, (-t), (t,))])
        '''x = fnn.Dense(features=self.boundary_hidden_dim, use_bias=False, name="shallow_fc1_indim"+str(2*t))(x)
        x = fnn.tanh(x)
        x = fnn.Dense(features=self.boundary_out_dim, use_bias=True, name="shallow_fc2_indim"+str(2*t))(x)
        return x'''

    # couldn't get this methods to work
    def default_model(self, features):
        
        model = partial(SNN_jax, u_history=self.u_history,
                                 u_window=self.u_window,
                                 p_window=self.p_window,
                                 out_dim=self.default_out_dim,
                                 hidden_dim=self.default_hidden_dim,
                                 n_layers=self.default_n_layers,
                                 dropout_prob=self.default_dropout_prob)
        return model(features)
        '''
        for i in range(self.default_n_layers - 1):
            x = jnp.hstack([x[self.u_history - self.u_window: self.u_history], x[-self.p_window:]])
            x = fnn.Dense(features=self.default_hidden_dim, use_bias=False, name=f"fc{i}")(x)
            x = self.scale*fnn.elu(x, alpha=self.alpha)
            x = Alpha_Dropout(rate=self.default_dropout_prob, deterministic=False)(x)
        x = fnn.Dense(features=self.default_out_dim, use_bias=True, name=f"fc{i + 1}")(x)
        return x'''
        
    @fnn.compact
    def __call__(self, t, features):
        if t <= self.num_boundary_models:
            return self.boundary_mod(features)
        else:
            return self.default_mod(features)
        '''self.default = functools.partial(SNN_jax, 
                                    self.u_history,
                                    self.u_window,
                                    self.p_window,
                                    self.default_out_dim,
                                    self.default_hidden_dim,
                                    self.default_n_layers,
                                    self.default_dropout_prob)

        self.boundary = functools.partial(ShallowBoundaryModel_jax, 
                                     self.u_history,
                                     self.u_history, # wasted parameters
                                     self.boundary_out_dim,
                                     self.boundary_hidden_dim)'''
        

        ''' output = jax.lax.cond(t <= self.num_boundary_models,
                              boundary,
                              default,
                              features)
            return output'''

        
        # t, features = t_features
        # model_idx = jnp.min(jnp.array([self.num_boundary_models, t]))
        # print('model_idx:' + str(model_idx))
        # return self.all_models[str(model_idx)](features)
        # return jax.lax.switch(model_idx, self.all_models, features)
        # model = jax.lax.dynamic_slice(np.array((self.all_models)), (t,), (1,))
        # return model(features)
        # return self.all_models[t](features)
        '''
        def true_func(x):
            return self.boundary_models[str(t)](x)
        def false_func(x):
            return self.default_model(x)
        
        output = jax.lax.cond(t <= self.num_boundary_models,
                              true_func,
                              false_func,
                              features)
        return output'''
        '''
        if t <= len(self.boundary_models):
            # return jax.lax.switch(t, self.boundary_models, features)
            return self.boundary_models[str(t)](features)
        else:
            return self.default_model(features)'''
        


import deluca


def convert_sim(sim):
    sim.__class__ = deluca.lung.environments.StitchedSim
    sim.inspiratory_model.__class__ = deluca.lung.utils.sim.nn.InspiratoryModel
    sim.inspiratory_model.default_model.__class__ = deluca.lung.utils.sim.nn.SNN

    for key in sim.inspiratory_model.boundary_dict:
        if sim.inspiratory_model.boundary_dict[key].__class__.__name__ == "ConstantModel":
            sim.inspiratory_model.boundary_dict[
                key
            ].__class__ = deluca.lung.utils.sim.nn.ConstantModel
        elif (
            sim.inspiratory_model.boundary_dict[key].__class__.__name__ == "RegressionBoundaryModel"
        ):
            sim.inspiratory_model.boundary_dict[
                key
            ].__class__ = deluca.lung.utils.sim.nn.RegressionBoundaryModel
        elif sim.inspiratory_model.boundary_dict[key].__class__.__name__ == "ShallowBoundaryModel":
            sim.inspiratory_model.boundary_dict[
                key
            ].__class__ = deluca.lung.utils.sim.nn.ShallowBoundaryModel
        elif sim.inspiratory_model.boundary_dict[key].__class__.__name__ == "SNN":
            sim.inspiratory_model.boundary_dict[key].__class__ = deluca.lung.utils.sim.nn.SNN
        else:
            print("ERROR!!!")

    u_mean, u_std = sim.u_scaler.mean, sim.u_scaler.std
    sim.u_scaler = deluca.lung.utils.core.TorchStandardScaler()
    sim.u_scaler.mean, sim.u_scaler.std = u_mean, u_std

    p_mean, p_std = sim.p_scaler.mean, sim.p_scaler.std
    sim.p_scaler = deluca.lung.utils.core.TorchStandardScaler()
    sim.p_scaler.mean, sim.p_scaler.std = p_mean, p_std

    return sim



import tqdm
import torch
import os

class AverageMeter(object):
    """Computes and stores the average and current value"""
    def __init__(self):
        self.reset()

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count

def train_simulator(
    sim,
    munger,
    u_window=5,
    p_window=3,
    device='cpu',
    exp_name='baseline',
    losscurve=False,
    train_key="train",
    test_key="test",
    batch_size=512,
    epochs=500,
    optimizer=torch.optim.Adam,
    optimizer_params={"lr": 1e-3},
    scheduler=torch.optim.lr_scheduler.ReduceLROnPlateau,
    scheduler_params={"factor": 0.9, "patience": 15},
    loss_fn=torch.nn.L1Loss,
    loss_fn_params={},
    print_loss=1,
):
    print("Getting data loader")
    loader = munger.get_data_loader(
        key=train_key, u_window=u_window, p_window=p_window, batch_size=batch_size
    )

    print("Scaling train data")
    X_tensor, y_tensor = munger.scale_and_window(
        key=train_key, u_window=u_window, p_window=p_window
    )

    print("Scaling test data")
    X_test_tensor, y_test_tensor = munger.scale_and_window(
        key=test_key, u_window=u_window, p_window=p_window
    )

    sim = sim.to(device)
    optim = optimizer(sim.parameters(), **optimizer_params)
    schedule = scheduler(optim, **scheduler_params)
    loss_fn = loss_fn(**loss_fn_params)

    

    # start training
    for epoch in tqdm.tqdm(range(epochs)):
        # training
        sim.reset()
        sim.train()
        train_objs = AverageMeter()
        for X_batch, y_batch in loader:
            X_batch = X_batch.to(device)
            y_batch = y_batch.to(device)

            sim.zero_grad()
            preds = sim.train_step(X_batch).squeeze()
            loss = loss_fn(preds, y_batch)
            loss.backward()
            optim.step()
            train_objs.update(loss.item(), batch_size)
        schedule.step(loss)

        # log the loss to tensorboard
        if losscurve:
            writer.add_scalar('train_loss', train_objs.avg, epoch)

        # testing
        if epoch % print_loss == 0:
            sim.reset()
            sim.eval()
            with torch.no_grad():
                X_tensor = X_tensor.to(device)
                y_tensor = y_tensor.to(device)
                X_test_tensor = X_test_tensor.to(device)
                y_test_tensor = y_test_tensor.to(device)
                preds = sim(X_tensor).squeeze()
                train_loss = loss_fn(preds, y_tensor)
                preds = sim(X_test_tensor).squeeze()
                test_loss = loss_fn(preds, y_test_tensor)
                print(
                    f"Epoch {epoch:2d}: train={train_loss.item():.5f}, test={test_loss.item():.5f}"
                )
            # log the loss to tensorboard
            if losscurve:
                writer.add_scalar('test_loss', test_loss, epoch)

    if losscurve:
        writer.flush()



import torch

from deluca.lung.controllers import PID
from deluca.lung.core import TorchController
from deluca.lung.utils import BreathWaveform
import itertools
import numpy as np
import random
from scipy.special import softmax
from deluca.lung.controllers import Expiratory


class DeepPIDResidual(TorchController):
    def __init__(
            self,
            H=100,
            waveform=None,
            bptt=1,
            input_dim=1,
            activation=torch.nn.ReLU,
            history_len=10,
            kernel_size=5,
            normalize=False,
            time_as_feature=False,
            u_scaler=None,
            p_scaler=None,
            pid_K=[3.0, 4.0, 0.0],
            **kwargs
    ):
        super().__init__()

        self.model = torch.nn.Sequential(
            torch.nn.Conv1d(input_dim, H, kernel_size),
            activation(),
            torch.nn.Flatten(),
            torch.nn.Linear(H * (history_len - kernel_size + 1), 1),
        )

        self.time_as_feature = time_as_feature
        if self.time_as_feature:
            self.time_model = torch.nn.Sequential(torch.nn.Linear(1, 10), activation(), torch.nn.Linear(10, 1))

        multiplier = self.tensor(1.0)
        self.residual_multiplier = torch.nn.Parameter(multiplier)

        # linear feature transform:
        # errs -> [average of last h errs, ..., average of last 2 errs, last err]
        # emulates low-pass filter bank

        self.featurizer = torch.ones((history_len, history_len), requires_grad=False).tril_()
        self.featurizer /= torch.arange(history_len, 0, -1).unsqueeze(0)

        self.history_len = history_len
        self.input_dim = input_dim
        self.waveform = waveform or BreathWaveform()
        self.pid_base = PID(K=pid_K)
        self.normalize = normalize
        if normalize:
            self.u_scaler = u_scaler
            self.p_scaler = p_scaler

        self.reset()

    def reset(self):
        self.errs = [self.tensor(0.0)] * self.history_len
        self.pid_base.waveform = self.waveform
        self.pid_base.reset()

    def update(self, key, state):
        getattr(self, key).append(state)

    def compute_action(self, state, t):
        self.pid_base.waveform = self.waveform
        target = self.tensor(self.waveform.at(t))
        if self.normalize:
            target_scaled = self.p_scaler.transform(target).squeeze()
            state_scaled = self.p_scaler.transform(state).squeeze()
            self.errs.append(target_scaled - state_scaled)
        else:
            self.errs.append(target - state)

        decay = self.waveform.decay(t)

        if decay is None:
            trajectory = torch.stack(self.errs[-self.history_len:]).unsqueeze(0).unsqueeze(0)
            u_in_base= self.pid_base.action(state, self.waveform.at(t),t)
            u_in_residual = self.model(trajectory @ self.featurizer)
            if self.time_as_feature:
                u_in_residual += self.time_model(self.tensor(self.cycle_phase(t)).unsqueeze(0))
            # u_in = u_in_base + u_in_residual * 0.1 * torch.relu(torch.tanh(5e-2 * self.residual_multiplier))
            u_in = u_in_base + u_in_residual * 1e-3
        else:
            u_in = self.tensor(decay)

        u_in = torch.clamp(u_in, min=0.0, max=100.0).squeeze()

        return u_in

    def train(
            self,
            sim,
            pip_feed="parallel",
            duration=3,
            dt=0.03,
            epochs=100,
            use_noise=False,
            optimizer=torch.optim.Adam,
            optimizer_params={"lr": 1e-3, "weight_decay": 1e-4},
            loss_fn=torch.nn.L1Loss,
            loss_fn_params={},
            scheduler=torch.optim.lr_scheduler.ReduceLROnPlateau,
            scheduler_params={"factor": 0.9, "patience": 10},
            use_tqdm=True,
            print_loss=1,
            shuffle=False,
            device="cpu",
    ):
        optimizer = optimizer(self.parameters(), **optimizer_params)
        scheduler = scheduler(optimizer, **scheduler_params)
        loss_fn = loss_fn(**loss_fn_params)

        tt = torch.linspace(0, duration, int(duration / dt))
        losses = []

        torch.autograd.set_detect_anomaly(True)

        PIPs = [10, 15, 20, 25, 30, 35]
        PEEP = 5

        # TODO: handle device-awareness
        for epoch in range(epochs):
            self.reset()
            sim.reset()

            if pip_feed == "parallel":
                self.zero_grad()
                loss = torch.tensor(0.0, device=device, requires_grad=True)

            for PIP in random.sample(PIPs, 6):

                if pip_feed == "sequential":
                    self.zero_grad()
                    loss = torch.tensor(0.0, device=device, requires_grad=True)

                self.waveform = BreathWaveform((PEEP, PIP))
                expiratory = Expiratory(waveform=self.waveform)

                for t in tt:
                    sim.pressure += use_noise * torch.normal(mean=torch.tensor(1.5), std=1.)
                    pressure = sim.pressure
                    u_in = self.compute_action(pressure, t)
                    u_out = expiratory.action(pressure, self.waveform.at(t), t)
                    sim.step(u_in, u_out) # potentially add multiplicative noise by * torch.normal(mean=torch.tensor(1.5), std=0.5)

                    if u_out == 0:
                        loss = loss + loss_fn(torch.tensor(self.waveform.at(t)), pressure)

                if pip_feed == "sequential":
                    loss.backward(retain_graph=True)
                    optimizer.step()
                    scheduler.step(loss)
                    per_step_loss = loss / len(tt)
                    losses.append(per_step_loss)

                    if epoch % print_loss == 0:
                        print(
                            f"Epoch: {epoch}, PIP: {PIP}\tLoss: {per_step_loss:.2f}\tLR: {optimizer.param_groups[0]['lr']}"
                        )

            if pip_feed == "parallel":
                loss.backward(retain_graph=True)
                optimizer.step()
                scheduler.step(loss)
                per_step_loss = loss / len(tt)
                losses.append(per_step_loss)

                if epoch % print_loss == 0:
                    print(
                        f"Epoch: {epoch}\tLoss: {per_step_loss:.2f}\tLR: {optimizer.param_groups[0]['lr']}"
                    )

        return losses

    def train_global(
            self,
            sims,
            pip_feed="parallel",
            duration=3,
            dt=0.03,
            epochs=100,
            use_noise=False,
            optimizer=torch.optim.Adam,
            optimizer_params={"lr": 1e-3, "weight_decay": 1e-4},
            loss_fn=torch.nn.L1Loss,
            loss_fn_params={},
            scheduler=torch.optim.lr_scheduler.ReduceLROnPlateau,
            scheduler_params={"factor": 0.9, "patience": 10},
            use_tqdm=True,
            print_loss=1,
            shuffle=False,
            device="cpu", 
        PIPs = [10, 15, 20, 25, 30, 35],
    ):
        optimizer = optimizer(self.parameters(), **optimizer_params)
        scheduler = scheduler(optimizer, **scheduler_params)
        loss_fn = loss_fn(**loss_fn_params)

        tt = torch.linspace(0, duration, int(duration / dt))
        losses = []

        torch.autograd.set_detect_anomaly(True)

#         PIPs = [10, 15, 20, 25, 30, 35]
        PEEP = 5

        # TODO: handle device-awareness
        for epoch in range(epochs):

            if pip_feed == "parallel":
                self.zero_grad()
                loss = torch.tensor(0.0, device=device, requires_grad=True)

            for PIP, sim in itertools.product(PIPs, sims):

                if pip_feed == "sequential":
                    self.zero_grad()
                    loss = torch.tensor(0.0, device=device, requires_grad=True)

                self.waveform = BreathWaveform((PEEP, PIP))

                self.reset()
                sim.reset()

                for t in tt:
                    sim.pressure += use_noise * torch.normal(mean=torch.tensor(1.5), std=1.)
                    pressure = sim.pressure
                    u_in, u_out = self(pressure, t)
                    sim(u_in, u_out,
                        t)  # potentially add multiplicative noise by * torch.normal(mean=torch.tensor(1.5), std=0.5)

                    if u_out == 0:
                        loss = loss + loss_fn(torch.tensor(self.waveform.at(t)), pressure)

                if pip_feed == "sequential":
                    loss.backward(retain_graph=True)
                    optimizer.step()
                    scheduler.step(loss)
                    per_step_loss = loss / len(tt)
                    losses.append(per_step_loss)

                    if epoch % print_loss == 0:
                        print(
                            f"Epoch: {epoch}, PIP: {PIP}\tLoss: {per_step_loss:.2f}\tLR: {optimizer.param_groups[0]['lr']}"
                        )

            if pip_feed == "parallel":
                loss.backward(retain_graph=True)
                optimizer.step()
                scheduler.step(loss)
                per_step_loss = loss / len(tt)
                losses.append(per_step_loss)

                if epoch % print_loss == 0:
                    print(
                        f"Epoch: {epoch}\tLoss: {per_step_loss:.2f}\tLR: {optimizer.param_groups[0]['lr']}"
                    )

        return losses

    def train_global_boosted(
            self,
            sim_datas, #Make this a list of (R, C, sim)s
            duration=3,
            dt=0.03,
            epochs=100,
            use_noise=False,
            optimizer=torch.optim.Adam,
            optimizer_params={"lr": 1e-3, "weight_decay": 1e-4},
            loss_fn=torch.nn.L1Loss,
            loss_fn_params={},
            scheduler=torch.optim.lr_scheduler.ReduceLROnPlateau,
            scheduler_params={"factor": 0.9, "patience": 1000},
            alpha=0.5,
            device="cpu",
    ):
        optimizer = optimizer(self.parameters(), **optimizer_params)
        scheduler = scheduler(optimizer, **scheduler_params)
        loss_fn = loss_fn(**loss_fn_params)

        tt = torch.linspace(0, duration, int(duration / dt))
        losses = []

        torch.autograd.set_detect_anomaly(True)

        PIPs = [10, 15, 20, 25, 30, 35]
        PEEP = 5
        PIPSims = list(itertools.product(PIPs, sim_datas))
        loss_by_simpip = np.zeros(len(PIPSims))
        self.zero_grad()

        idxs =range(len(PIPSims))
        for epoch in range(epochs):
            p = softmax(alpha * loss_by_simpip)
            loss = torch.tensor(0.0, device=device, requires_grad=True)
            for idx in idxs:
                PIP, (R, C, sim) = PIPSims[idx]
                self.waveform = BreathWaveform((PEEP, PIP))

                self.reset()
                sim.reset()
                loss_idx = 0.0
                for t in tt:
                    sim.pressure += use_noise * torch.normal(mean=torch.tensor(1.5), std=1.)
                    pressure = sim.pressure
                    u_in, u_out = self(pressure, t)
                    sim(u_in, u_out, t)

                    if u_out == 0:
                        curr_loss = loss_fn(torch.tensor(self.waveform.at(t)), pressure)
                        loss = loss + curr_loss * p[idx]
                        loss_idx += curr_loss

                loss.backward(retain_graph=True)
                per_step_loss = loss_idx / len(tt)
                losses.append(per_step_loss)
                loss_by_simpip[idx] = per_step_loss
            optimizer.step()
            self.zero_grad()

            print(f"Epoch: {epoch}, PIP: {PIP} R: {R} C: {C}\tLoss: {per_step_loss:.2f}\tLR: {optimizer.param_groups[0]['lr']}")
            for i in range(len(PIPSims)):
                PIP,(R,C, _) = PIPSims[i]
                print(f"PIP, {PIP}, R {R}, C {C}, loss {loss_by_simpip[i] :.2f}")
            print(f"loss mean {np.mean(loss_by_simpip):.2f}+-{np.std(loss_by_simpip):.2f}")
        return losses



import jax

from deluca.lung.core import Controller


class PIDJax(Controller):
    def __init__(self, K=[3., 4., 0.], RC=0.5, jaxed=False, **kwargs):
        self.init_K = K
        self.RC = RC
        self.reset()

    def reset(self):
        self.coef = [0., 0., 0.]

    def action(self, pressure, target, t):
        err = target - pressure
        decay = self.dt / (self.dt + self.RC)

        # self.coef += 



import torch
import random
from deluca.lung.controllers import Expiratory
from deluca.lung.utils import BreathWaveform
import jax
import jax.numpy as jnp
import optax

def rollout(controller, sim, waveform, tt, use_noise, loss_fn, loss):
    expiratory = Expiratory.create(waveform=waveform)

    controller_state = controller.init()
    expiratory_state = expiratory.init()
    sim_state, obs = sim.reset() 
    def loop_over_tt(ctrlState_expState_simState_obs_loss, t):
        controller_state, expiratory_state, sim_state, obs, loss = ctrlState_expState_simState_obs_loss
        mean = 1.0
        std = 1.0
        noise = mean + std * jax.random.normal(jax.random.PRNGKey(0), shape=())
        pressure = sim_state.predicted_pressure + use_noise * noise
        sim_state = sim_state.replace(predicted_pressure=pressure) # Need to update p_history as well or no?
        obs = obs.replace(predicted_pressure=pressure, time=t)

        controller_state, u_in = controller(controller_state, obs)
        expiratory_state, u_out = expiratory(expiratory_state, obs)

        sim_state, obs = sim(sim_state, (u_in, u_out))
        
        loss = jax.lax.cond(u_out == 0,
                            lambda x: x + loss_fn(jnp.array(waveform.at(t)), pressure),
                            lambda x: x,
                            loss)
        return (controller_state, expiratory_state, sim_state, obs, loss), None
    (_, _, _, _, loss), _ = jax.lax.scan(loop_over_tt, (controller_state, expiratory_state, sim_state, obs, loss), tt)
    return loss

# TODO: add scheduler and scheduler_params
# Question: Jax analogue of torch.autograd.set_detect_anomaly(True)?
def train_controller_jax(
    controller,
    sim,
    waveform,
    duration=3,
    dt=0.03,
    epochs=100,
    use_noise=False,
    optimizer=optax.adamw,
    optimizer_params={"learning_rate": 1e-3, "weight_decay": 1e-4},
    loss_fn=lambda x, y: (jnp.abs(x - y)).mean(),
    # scheduler=torch.optim.lr_scheduler.ReduceLROnPlateau,
    # scheduler_params={},
    use_tqdm=True,
    print_loss=1,
    shuffle=False,
):
    optim = optimizer(**optimizer_params)
    optim_state = optim.init(controller)

    tt = jnp.linspace(0, duration, int(duration / dt))
    losses = []

    # torch.autograd.set_detect_anomaly(True)

    for epoch in range(epochs):
        value, grad = jax.value_and_grad(rollout)(controller, sim, waveform, tt, use_noise, loss_fn, jnp.array(0.))
        updates, optim_state = optim.update(grad, optim_state, controller)
        controller = optax.apply_updates(controller, updates)
        per_step_loss = value / len(tt)
        losses.append(per_step_loss)

        if epoch % print_loss == 0:
            print(
                f"Epoch: {epoch}\tLoss: {per_step_loss:.2f}"
            )
    return controller
    

def train_controller(
    controller,
    sim,
    waveform,
    duration=3,
    dt=0.03,
    epochs=100,
    use_noise=False,
    optimizer=torch.optim.Adam,
    optimizer_params={"lr": 1e-3, "weight_decay": 1e-4},
    loss_fn=torch.nn.L1Loss,
    loss_fn_params={},
    scheduler=torch.optim.lr_scheduler.ReduceLROnPlateau,
    scheduler_params={},
    use_tqdm=True,
    print_loss=1,
    shuffle=False,
    device="cpu",
):
    optimizer = optimizer(controller.parameters(), **optimizer_params)
    scheduler = scheduler(optimizer, **scheduler_params)
    loss_fn = loss_fn(**loss_fn_params)

    tt = torch.linspace(0, duration, int(duration / dt))
    losses = []

    torch.autograd.set_detect_anomaly(True)

    # TODO: handle device-awareness
    for epoch in range(epochs):
        controller.reset()
        sim.reset()

        loss = torch.tensor(0.0, device=device, requires_grad=True)

        controller.zero_grad()

        expiratory = Expiratory(waveform=waveform)

        for t in tt:
            sim.pressure += use_noise * torch.normal(mean=torch.tensor(1.0), std=1.0)
            pressure = sim.pressure
            u_in = controller.action(pressure, waveform.at(t), t)
            u_out = expiratory.action(pressure, waveform.at(t), t)
            sim.step(u_in, u_out)

            if u_out == 0:
                loss = loss + loss_fn(torch.tensor(waveform.at(t)), pressure)

        loss.backward(retain_graph=True)
        optimizer.step()
        scheduler.step(loss)
        per_step_loss = loss / len(tt)
        losses.append(per_step_loss)
        if epoch % print_loss == 0:
            print(
                f"Epoch: {epoch}\tLoss: {per_step_loss:.2f}\tLR: {optimizer.param_groups[0]['lr']}"
            )

    return losses


def train_controller_multipip(
    controller,
    sim,
    pip_feed="sequential",
    duration=3,
    dt=0.03,
    epochs=100,
    use_noise=False,
    optimizer=torch.optim.Adam,
    optimizer_params={"lr": 1e-3, "weight_decay": 1e-4},
    loss_fn=torch.nn.L1Loss,
    loss_fn_params={},
    scheduler=torch.optim.lr_scheduler.ReduceLROnPlateau,
    scheduler_params={"factor": 0.9, "patience": 10},
    use_tqdm=True,
    print_loss=1,
    shuffle=False,
    device="cpu",
):
    optimizer = optimizer(controller.parameters(), **optimizer_params)
    scheduler = scheduler(optimizer, **scheduler_params)
    loss_fn = loss_fn(**loss_fn_params)

    tt = torch.linspace(0, duration, int(duration / dt))
    losses = []

    torch.autograd.set_detect_anomaly(True)

    PIPs = [10, 15, 20, 25, 30, 35]
    PEEP = 5

    # TODO: handle device-awareness
    for epoch in range(epochs):
        controller.reset()
        sim.reset()

        if pip_feed == "parallel":
            controller.zero_grad()
            loss = torch.tensor(0.0, device=device, requires_grad=True)

        for PIP in random.sample(PIPs, 6):

            if pip_feed == "sequential":
                controller.zero_grad()
                loss = torch.tensor(0.0, device=device, requires_grad=True)

            waveform = BreathWaveform((PEEP, PIP))
            expiratory = Expiratory(waveform=waveform)

            for t in tt:
                sim.pressure += use_noise * torch.normal(
                    mean=torch.tensor(1.5), std=1.0
                )
                pressure = sim.pressure
                u_in = controller.action(pressure, waveform.at(t), t)
                u_out = expiratory.action(pressure, waveform.at(t), t)
                sim.step(u_in, u_out)

                if u_out == 0:
                    loss = loss + loss_fn(torch.tensor(waveform.at(t)), pressure)

            if pip_feed == "sequential":
                loss.backward(retain_graph=True)
                optimizer.step()
                scheduler.step(loss)
                per_step_loss = loss / len(tt)
                losses.append(per_step_loss)

                if epoch % print_loss == 0:
                    print(
                        f"Epoch: {epoch}, PIP: {PIP}\tLoss: {per_step_loss:.2f}\tLR: {optimizer.param_groups[0]['lr']}"
                    )

        if pip_feed == "parallel":
            loss.backward(retain_graph=True)
            optimizer.step()
            scheduler.step(loss)
            per_step_loss = loss / len(tt)
            losses.append(per_step_loss)

            if epoch % print_loss == 0:
                print(
                    f"Epoch: {epoch}\tLoss: {per_step_loss:.2f}\tLR: {optimizer.param_groups[0]['lr']}"
                )

    return losses



import os
from functools import partial
from enum import Enum
import jax
import jax.numpy as jnp
import deluca.core

DEFAULT_PRESSURE_RANGE = (5.0, 35.0)
DEFAULT_KEYPOINTS = [1e-8, 1.0, 1.5, 3.0]
DEFAULT_BPM = 20


class Phase(Enum):
    RAMP_UP = 1
    PIP = 2
    RAMP_DOWN = 3
    PEEP = 4


class BreathWaveform(deluca.Obj):
    """Waveform generator with shape |‾\_"""

    custom_range: tuple = deluca.field(jaxed=False)
    lo: float = deluca.field(jaxed=False)
    hi: float = deluca.field(jaxed=False)
    fp: jnp.array = deluca.field(jaxed=False)
    xp: jnp.array = deluca.field(jaxed=False)
    dtype: jax._src.numpy.lax_numpy._ScalarMeta = deluca.field(jnp.float32, jaxed=False)
    keypoints: jnp.array = deluca.field(jaxed=False)
    bpm: int = deluca.field(DEFAULT_BPM, jaxed=False)
    kernel: list = deluca.field(jaxed=False)
    dt: float = deluca.field(0.01, jaxed=False)
    period: float = deluca.field(jaxed=False)

    def setup(self):
        self.lo, self.hi = self.custom_range or DEFAULT_PRESSURE_RANGE
        self.fp = jnp.array([self.lo, self.hi, self.hi, self.lo, self.lo])

        self.xp = jnp.zeros(len(self.fp), dtype=self.dtype)
        if self.keypoints:
            self.xp = jax.ops.index_update(
                self.xp, jax.ops.index[1:], jnp.array(self.keypoints, dtype=self.dtype)
            )
        else:
            self.xp = jax.ops.index_update(
                self.xp, jax.ops.index[1:], jnp.array(DEFAULT_KEYPOINTS, dtype=self.dtype)
            )
        self.xp = jax.ops.index_update(self.xp, -1, 60 / self.bpm)
        self.keypoints = self.xp
        self.keypoints = self.xp
        self.period = float(self.xp[-1])

        pad = 0
        num = int(1 / self.dt)
        if self.kernel is not None:
            pad = 60 / bpm / (num - 1)
            num += len(kernel) // 2 * 2

        tt = jnp.linspace(-pad, 60 / self.bpm + pad, num, dtype=self.dtype)
        self.fp = self.at(tt)
        self.xp = tt

        if self.kernel is not None:
            self.fp = jnp.convolve(self.fp, self.kernel, mode="valid")
            self.xp = jnp.linspace(0, 60 / bpm, int(1 / self.dt), dtype=dtype)

    """
    @property
    def keypoints(self):
        if hasattr(self, "_keypoints"):
            return self._keypoints
        else:
            return self.xp"""
    """
    @property
    def period(self):
        return self.xp[-1]"""

    @property
    def PIP(self):
        if hasattr(self, "custom_range"):
            return self.custom_range[1]
        else:
            return jnp.max(self.fp)

    @property
    def PEEP(self):
        if hasattr(self, "custom_range"):
            return self.custom_range[0]
        else:
            return jnp.min(self.fp)

    def is_in(self, t):
        return self.elapsed(t) <= self.keypoints[2]

    def is_ex(self, t):
        return not self.is_in(t)

    def at(self, t):
        @partial(jax.jit, static_argnums=(3,))
        def static_interp(t, xp, fp, period):
            return jnp.interp(t, xp, fp, period=period)

        return static_interp(t, self.xp, self.fp, self.period).astype(self.dtype)

    def elapsed(self, t):
        return t % self.period

    # TODO: change all files where "if decay is None" to "if decay == float("inf")"
    # files affected: _mpc, _clipped_adv_deep, _clipped_deep, _deep_pid_residual,
    # _deep_pid_residual_clipped
    def decay(self, t):
        elapsed = self.elapsed(t)

        def false_func():
            result = jax.lax.cond(
                elapsed < self.keypoints[3],
                lambda x: 0.0,
                lambda x: 5 * (1 - jnp.exp(5 * (self.keypoints[3] - elapsed))).astype(self.dtype),
                None,
            )
            return result

        # float(inf) as substitute to None since cond requries same type output
        result = jax.lax.cond(
            elapsed < self.keypoints[2], lambda x: float("inf"), lambda x: false_func(), None
        )
        return result

    def phase(self, t):
        return jnp.searchsorted(self.keypoints, jnp.mod(t, self.period), side="right")



import os
import pickle

import torch
import numpy as np
import matplotlib.pyplot as plt

from deluca.lung.controllers import ResidualExplorer, PID
from deluca.lung.environments import PhysicalLung
from deluca.lung.utils.data.analyzer import Analyzer
from deluca.lung.utils.core import BreathWaveform
from deluca.lung.utils.scripts.run_calibration import run_calibration
from deluca.lung.utils.scripts.run_controller import run_controller


plt.rc("figure", figsize=(10, 3))


def collect_runs(controller, path, n_runs=1, append_to=None, env=None, **kwargs):
    if append_to is None:
        results = []
    else:
        results = append_to

    for i in range(n_runs):
        env.reset()
        result = run_controller(controller, env=env, **kwargs)
        results.append(result)
        pickle.dump(result, open(path % i, "wb"))
    return results


def run_explorer(
    controller,
    directory,
    R=50,
    C=10,
    PEEP=5,
    T=10000,
    n_runs=5,
    PIPs=[10, 15, 20, 25, 30, 35],
    abort=70,
    dt=0.03,
    env=None,
):
    if not isinstance(controller, ResidualExplorer):
        print("ERROR: expecting ResidualExplorer as controller")
        return

    env = env or PhysicalLung()
    print("Running calibration")
    run_calibration(R, C, PEEP, directory)

    all_results = {}
    for PIP in PIPs:
        path = os.path.join(directory, f"R{R}C{C}PEEP{PEEP}_PIP{PIP}_pid_triangle_residual_%i.pkl")

        controller.waveform = BreathWaveform((PEEP, PIP))

        results = collect_runs(
            controller,
            path,
            R=R,
            C=C,
            PEEP=PEEP,
            T=T,
            dt=dt,
            n_runs=n_runs,
            abort=abort,
            env=env,
        )
        all_results[PIP] = results

    print("Running calibration")
    run_calibration(R, C, PEEP, directory)

    return all_results



from functools import partial

import jax
import jax.numpy as jnp
import flax.linen as fnn
import optax

import deluca.core
from deluca.lung.core import LungEnv
from deluca.lung.utils.data.transform import ShiftScaleTransform

from deluca.lung.utils.sim.nn import SNN
from deluca.lung.utils.sim.nn import ShallowBoundaryModel

class StitchedSimObservation(deluca.Obj):
    predicted_pressure: float = 0.0
    time: float = 0.0

class SimulatorState(deluca.Obj):
    u_history: jnp.ndarray
    p_history: jnp.ndarray
    t_in: int = 0
    steps: int = 0
    predicted_pressure: float = 0.0


class StitchedSim(LungEnv):
    params: list = deluca.field(jaxed=True)

    init_rng: jnp.array = deluca.field(jaxed=False)
    u_window: int = deluca.field(5, jaxed=False)
    p_window: int = deluca.field(3, jaxed=False)
    u_history_len: int = deluca.field(5, jaxed=False)
    p_history_len: int = deluca.field(5, jaxed=False)
    u_scaler: ShiftScaleTransform = deluca.field(jaxed=False)
    p_scaler: ShiftScaleTransform = deluca.field(jaxed=False)
    to_round: bool = deluca.field(False, jaxed=False)
    seed: int = deluca.field(0, jaxed=False)
    flow: int = deluca.field(0, jaxed=False)

    default_out_dim: int = deluca.field(1, jaxed=False)
    default_hidden_dim: int = deluca.field(100, jaxed=False)
    default_n_layers: int = deluca.field(4, jaxed=False)
    default_dropout_prob: float = deluca.field(0.0, jaxed=False)

    num_boundary_models: int = deluca.field(5, jaxed=False)
    boundary_out_dim: int = deluca.field(1, jaxed=False)
    boundary_hidden_dim: int = deluca.field(100, jaxed=False)
    reset_scaled_peep: float = deluca.field(0.0, jaxed=False)

    default_model: fnn.module = deluca.field(SNN, jaxed=False)
    boundary_models: list = deluca.field(default_factory=list, jaxed=False)
    ensemble_models: list = deluca.field(default_factory=list, jaxed=False)

    def setup(self):
        self.u_history_len = max(self.u_window, self.num_boundary_models)
        self.p_history_len = max(self.p_window, self.num_boundary_models)

        self.default_model = SNN(
            out_dim=self.default_out_dim,
            hidden_dim=self.default_hidden_dim,
            n_layers=self.default_n_layers,
        )
        default_params = self.default_model.init(
            jax.random.PRNGKey(0), jnp.ones([self.u_history_len + self.p_history_len])
        )["params"]

        self.boundary_models = [
            ShallowBoundaryModel(
                out_dim=self.boundary_out_dim, hidden_dim=self.boundary_hidden_dim, model_num=i + 1
            )
            for i in range(self.num_boundary_models)
        ]
        boundary_params = [
            self.boundary_models[i].init(
                jax.random.PRNGKey(0), jnp.ones([self.u_history_len + self.p_history_len])
            )["params"]
            for i in range(self.num_boundary_models)
        ]

        self.ensemble_models = self.boundary_models + [self.default_model]
        # print("ENSEMBLE MODELS:")
        # print(self.ensemble_models)
        # without ensemble i.e. flattened params
        if self.params is None:
            self.params = boundary_params + [default_params]
        # print("TREE MAP")
        # print(jax.tree_map(lambda x: x.shape, self.params))

    def reset(self):
        scaled_peep = self.reset_scaled_peep
        state = SimulatorState(
            u_history=jnp.zeros([self.u_history_len]),
            p_history=jnp.hstack([jnp.zeros([self.p_history_len]), jnp.array([scaled_peep])]),
            t_in=0,
            steps=0,
            predicted_pressure=scaled_peep,
        )
        state = self.sync_unscaled_pressure(state)
        obs = StitchedSimObservation(predicted_pressure=state.predicted_pressure, time=self.time(state))
        return state, obs

    def update_history(self, history, value):
        history = jnp.roll(history, shift=-1)
        history = history.at[-1].set(value)
        return history

    def sync_unscaled_pressure(self, state):
        scaled_pressure = self.p_scaler.inverse(state.predicted_pressure).squeeze()
        return state.replace(predicted_pressure=scaled_pressure)

    def __call__(self, state, action):
        u_in, u_out = action
        model_idx = jnp.min(jnp.array([state.t_in, self.num_boundary_models]))
        funcs = [partial(self.ensemble_models[i].apply, {"params": self.params[i]}) 
                 for i in range(self.num_boundary_models+1)]
        def true_func(state):
            def call_if_t_in_positive(state):
                new_state, _ = self.reset()
                return new_state

            state = jax.lax.cond(state.t_in > 0, call_if_t_in_positive, lambda x: x, state)
            return state

        def false_func(state, u_in, model_idx):
            if self.to_round:
                u_in_scaled = self.u_scaler(jnp.round(u_in)).squeeze()
            else:
                u_in_scaled = self.u_scaler(u_in).squeeze()
            state = state.replace(
                u_history=self.update_history(state.u_history, u_in_scaled), t_in=state.t_in + 1
            )
            boundary_features = jnp.hstack(
                [state.u_history[-self.u_history_len :], state.p_history[-self.p_history_len :]]
            )
            default_features = jnp.hstack(
                [state.u_history[-self.u_window :], state.p_history[-self.p_window :]]
            )
            default_pad_len = (self.u_history_len + self.p_history_len) - (
                self.u_window + self.p_window
            )
            default_features = jnp.hstack([jnp.zeros((default_pad_len,)), default_features])
            features = jax.lax.cond(
                model_idx == self.num_boundary_models,
                lambda x: default_features,
                lambda x: boundary_features,
                None,
            )

            scaled_pressure = jax.lax.switch(model_idx, funcs, features)[0]

            state = state.replace(predicted_pressure=scaled_pressure.squeeze())
            new_p_history = self.update_history(state.p_history, scaled_pressure.squeeze())
            state = state.replace(
                p_history=new_p_history
            )  # just for inference. This is undone during training
            state = self.sync_unscaled_pressure(state)  # unscale predicted pressure
            return state

        partial_false_func = partial(false_func, u_in=u_in, model_idx=model_idx)
        state = jax.lax.cond(u_out == 1, true_func, partial_false_func, state)
        obs = StitchedSimObservation(predicted_pressure=state.predicted_pressure, time=self.time(state))
        return state.replace(steps=state.steps + 1), obs


def get_X_y_for_next_epoch_tf(loader, batch_size):
    # loader is a tensorflow.data.Dataset
    loader = loader.shuffle(buffer_size=len(loader))
    loader = loader.batch(batch_size=batch_size, drop_remainder=True)
    loader_list = list(loader)
    unzipped_loader = list(zip(*loader_list))
    X = jnp.array([jnp.array(z.numpy()) for z in unzipped_loader[0]])
    y = jnp.array([jnp.array(z.numpy()) for z in unzipped_loader[1]])
    return X, y


def rollout(model, data, num_boundary_models):
    # data.shape == (2, N)
    start_idx = 0
    end_idx = len(data[0])
    state, _ = model.reset()
    new_u_history = jnp.zeros((num_boundary_models,))
    new_p_history = jnp.zeros((num_boundary_models,))
    state = state.replace(u_history=new_u_history, p_history=new_p_history)
    state_loss_init = (state, 0.0)

    def predict_and_update_state(i, state_loss):
        state, loss = state_loss
        u_in, pressure = data[0, i], data[1, i]  # unscaled u_in and pressure
        new_p_history = jnp.append(state.p_history, model.p_scaler(pressure).squeeze())
        state = state.replace(p_history=new_p_history)
        next_state, _ = model(state=state, action=(u_in, 0))
        # drop predicted_pressure so that u_history.shape = u_history_len+1, p_history.shape = p_history_len+1
        truncated_p_history = next_state.p_history[:-1]
        next_state = next_state.replace(p_history=truncated_p_history)
        pred = model.p_scaler(next_state.predicted_pressure)  # scaled gradient step
        return (
            next_state,
            loss + jnp.abs(model.p_scaler(data[1, i + 1]) - pred),
        )  # scaled gradient step

    (state, total_loss) = jax.lax.fori_loop(
        start_idx, end_idx, predict_and_update_state, state_loss_init
    )
    return total_loss / (end_idx - start_idx)


def loop_over_loader(model_optimState, X_Y, optim, rollout):
    """
    rollout has signature (model, data) -> loss where data.shape = (2, N)
    X_batch.shape = Y_batch.shape = (num_batches, batch_size, N=29)
    """
    # print('=================== ENTER loop_over_loader ======================')
    X_batch, y_batch = X_Y
    model, optim_state = model_optimState
    value, grad = jax.value_and_grad(map_rollout_over_batch)(model, (X_batch, y_batch), rollout)
    updates, optim_state = optim.update(grad, optim_state, model)
    model = optax.apply_updates(model, updates)
    return (model, optim_state), None


def map_rollout_over_batch(model, data, rollout):
    """
    rollout has signature (model, data) -> loss where data.shape = (2, N)
    data.shape = ((batch_size, N), (batch_size, N))
    """
    rollout_partial = lambda xs: partial(rollout, model=model)(data=xs)
    data_zipped = jnp.array(list(zip(data[0], data[1])))  # (batch_size, 2, N)
    losses = jax.vmap(rollout_partial)(data_zipped)
    return jnp.array(losses).mean()


# TODO: add scheduler and scheduler_params
def stitched_sim_train(
    munger,
    model,
    num_boundary_models,
    # 0 to num_boundary_models-1 are boundary models, and num_boundary_models is #
    # default_model
    train_key="train",
    test_key="test",
    batch_size=512,
    epochs=500,
    optimizer=optax.adam,
    optimizer_params={"learning_rate": 1e-3},
    # scheduler=torch.optim.lr_scheduler.ReduceLROnPlateau,
    # scheduler_params={"factor": 0.9, "patience": 15},
    loss_fn=lambda x, y: (jnp.abs(x - y)).mean(),
    # loss_fn_params={},
    print_loss=100,
):
    # Unpack dataloader
    loader = munger.get_dataloader_tf(
        key=train_key, shuffle=True, scale_data=False, batch_size=batch_size
    )

    # evaluate on these at end of epoch
    X_train, y_train = munger.get_scaled_data_jax(train_key, scale_data=False)
    X_test, y_test = munger.get_scaled_data_jax(test_key, scale_data=False)

    # set up model and optimizer
    optim = optimizer(**optimizer_params)
    optim_state = optim.init(model)

    rollout_partial = partial(rollout, num_boundary_models=num_boundary_models)
    loop_over_loader_partial = partial(loop_over_loader, optim=optim, rollout=rollout_partial)
    for epoch in range(epochs + 1):
        if epoch % 25 == 0:
            print("epoch:" + str(epoch))
        X, y = get_X_y_for_next_epoch_tf(loader, batch_size)

        (model, optim_state), _ = jax.lax.scan(
            loop_over_loader_partial, (model, optim_state), (X, y)
        )

        if epoch % print_loss == 0:
            # expensive end-of-epoch eval, just for intuition
            print("X_train.shape:" + str(X_train.shape))
            print("y_train.shape:" + str(y_train.shape))
            train_loss = map_rollout_over_batch(model, (X_train, y_train), rollout_partial)
            # cross-validation
            print("X_test.shape:" + str(X_test.shape))
            print("y_test.shape:" + str(y_test.shape))
            test_loss = map_rollout_over_batch(model, (X_test, y_test), rollout_partial)
            print(f"Epoch {epoch:2d}: train={train_loss.item():.5f}, test={test_loss.item():.5f}")
            print("-----------------------------------")
    print("finished looping over epochs")
    return model




##########################################################
###################
################### SingleComp Lung Environment
###################
##########################################################

import torch
import numpy as np

from deluca.lung.core import LungEnv

# Sources:
# https://github.com/CohenLabPrinceton/Ventilator-Dev/blob/master/sandbox/HOWTO_RunController.ipynb
# https://github.com/CohenLabPrinceton/Ventilator-Dev/blob/master/vent/controller/control_module.py
# https://github.com/MinRegret/venti
def PropValve(x):  # copied from Controller.__SimulatedPropValve
    y = 3 * x
    y_torch = torch.from_numpy(np.array([0.03 * (y - 130)])).float()
    flow_new = 1.0 * (torch.tanh(y_torch) + 1)

    return torch.clamp(flow_new, 0.0, 1.72)


def Solenoid(x):  # copied from Controller.__SimulatedSolenoid
    if x > 0:
        return x / x
    else:
        return x * 0.0


class SingleCompLung(LungEnv):
    def __init__(self, resistance=15, compliance=50, min_volume=0.2, peep_value=5.0, **kwargs):
        # dynamics hyperparameters
        self.min_volume = min_volume
        self.P0 = 10.0
        self.peep_value = peep_value
        self.R = resistance
        self.C = compliance
        self.flow = 0.0

        # reset states
        self.reset()

    def reset(self):
        # volume is the only free parameter
        self.volume = self.min_volume
        self.pressure = (self.flow * self.R) + (self.volume / self.C) + self.P0

    def compute_aux_states(self):
        self.pressure = (self.flow * self.R) + (self.volume / self.C) + self.P0

    def step(self, u_in, u_out):
        dt = self.dt

        # calculate flow from u_in
        self.flow = torch.clamp(PropValve(u_in), 0, 2)
        # update volume by flow rate
        self.volume = self.volume + self.flow * dt

        # compute and record state
        self.compute_aux_states()

        return self.pressure



import jax
import jax.numpy as jnp
import deluca
from functools import partial
from typing import List, Callable
from deluca.lung.core import Controller
from deluca.lung.core import DEFAULT_DT
from deluca.lung.core import proper_time


class Predestined(Controller):
    time: float = deluca.field(jaxed=False)
    steps: int = deluca.field(jaxed=False)
    dt: float = deluca.field(jaxed=False)
    u_ins: jnp.array = deluca.field(jaxed=False)
    switch_funcs: List[Callable] = deluca.field(jaxed=False)
    def setup(self):
        self.switch_funcs = [partial(lambda x, y: self.u_ins[y], y=i) for i in range(len(self.u_ins))]

    def __call__(self, state, obs, *args, **kwargs):
        action = jax.lax.switch(state.steps, self.switch_funcs, None)
        time = obs.time
        new_dt = jnp.max(jnp.array([DEFAULT_DT, time - proper_time(state.time)]))
        new_time = time
        new_steps = state.steps + 1
        state = state.replace(time=new_time, steps=new_steps, dt=new_dt)
        return state, action



from deluca.lung.core import Controller


class Impulse(Controller):
    def __init__(self, impulse=50, start=0.5, end=0.65):
        self.impulse = impulse
        self.start = start
        self.end = end

    def action(self, pressure, target, t):
        impulse = 0
        if self.time >= self.start and self.time <= self.end:
            impulse = self.impulse

        return impulse



""" Module for interacting with physical and/or simulated devices installed on the ventilator.

"""

import os
from importlib import import_module
from ast import literal_eval
from deluca.lung.devices.sensors import Sensor
from deluca.lung.devices.base import PigpioConnection

import deluca.lung.devices.valves as valves
import configparser


default_hal_config = os.path.join(os.path.abspath(os.path.dirname(__file__)), "config/devices.ini")


class Hal:
    """Hardware Abstraction Layer for ventilator hardware.
    Defines a common API for interacting with the sensors & actuators on the ventilator. The types of devices installed
    on the ventilator (real or simulated) are specified in a configuration file.
    """

    def __init__(self, config_file=default_hal_config):
        """Initializes HAL from config_file.
            For each section in config_file, imports the class <type> from module <module>, and sets attribute
            self.<section> = <type>(**opts), where opts is a dict containing all of the options in <section> that are
            not <type> or <section>. For example, upon encountering the following entry in config_file.ini:

                [adc]
                type   = ADS1115
                module = devices
                i2c_address = 0x48
                i2c_bus = 1

            The Hal will:
                1) import deluca.lung.devices.ADS1115 (or ADS1015) as a local variable:
                        class_ = getattr(import_module('.devices', 'vent.io'), 'ADS1115')

                2) Instantiate an ADS1115 object with the arguments defined in config_file and set it as an attribute:
                        self._adc = class_(gpio=self._gpio,address=0x48,i2c_bus=1)

            Note: RawConfigParser.optionxform() is overloaded here s.t. options are case sensitive (they are by default
            case insensitive). This is necessary due to the kwarg MUX which is so named for consistency with the config
            registry documentation in the ADS1115 datasheet. For example, A P4vMini pressure_sensor on pin A0 (MUX=0)
            of the ADC is passed arguments like:

            analog_sensor = AnalogSensor(
                _ig=self._gpio,
                adc=self._adc,
                MUX=0,
                offset_voltage=0.25,
                output_span = 4.0,
                conversion_factor=2.54*20
            )

            Note: ast.literal_eval(opt) interprets integers, 0xFF, (a, b) etc. correctly. It does not interpret strings
            correctly, nor does it know 'adc' -> self._adc; therefore, these special cases are explicitly handled.
        Args:
            config_file (str): Path to the configuration file containing the definitions of specific components on the
                ventilator machine. (e.g., config_file = "vent/io/config/devices.ini")
        """
        self._setpoint_in = 0.0  # setpoint for inspiratory side
        self._setpoint_ex = 0.0  # setpoint for expiratory side
        self._adc = object
        self._control_valve = object
        self._expiratory_valve = object
        self._pressure_sensor = object
        self._flow_sensor_ex = object
        self._gpio = object
        self.config = configparser.RawConfigParser()
        self.config.optionxform = lambda option: option
        self.config.read(config_file)

        for section in self.config.sections():
            sdict = dict(self.config[section])
            class_ = getattr(import_module("." + sdict["module"], "deluca.lung"), sdict["type"])
            opts = {
                key: sdict[key]
                for key in sdict.keys()
                - (
                    "module",
                    "type",
                )
            }
            for key in opts.keys():
                if key == "adc":
                    opts[key] = self._adc
                else:
                    opts[key] = literal_eval(opts[key])
            print(
                "  [ {device_name:^19} ]  opts: {device_options}".format(
                    device_name=section, device_options=opts
                )
            )  # debug
            if section == "gpio":
                setattr(self, "_gpio", class_(**opts))
                # Clean up i2c
                for i in range(32):
                    try:
                        self._gpio.i2c_close(i)
                    except:
                        continue
            else:
                setattr(self, "_" + section, class_(gpio=self._gpio, **opts))


    @property
    def pressure(self) -> float:
        """Returns the pressure from the primary pressure sensor."""
        return self._pressure_sensor.get()

    @property
    def flow_ex(self) -> float:
        """ The measured flow rate expiratory side."""
        return self._flow_sensor_ex.get()

    @property
    def setpoint_in(self) -> float:
        """ The currently requested flow for the inspiratory proportional control valve as a proportion of maximum."""
        return self._setpoint_in

    @setpoint_in.setter
    def setpoint_in(self, value: float):
        """Sets the openness of the inspiratory valve to the requested value.

        Args:
            value: Requested flow, as a proportion of maximum. Must be in [0, 1].
        """
        if not 0 <= value <= 100:
            raise ValueError("setpoint must be a number between 0 and 100")
        self._control_valve.setpoint = value
        self._setpoint_in = value

    @property
    def setpoint_ex(self) -> float:
        """ The currently requested flow on the expiratory side as a proportion of the maximum."""
        return self._setpoint_ex

    @setpoint_ex.setter
    def setpoint_ex(self, value):
        """Sets the openness of the expiratory valve to the requested value.

        Args:
            value (float): Requested flow, as a proportion of maximum. Must be either 0 or 1 for OnOffValve, and between
                0 and 1 for a (proportional) control valve.
        """
        if isinstance(self._expiratory_valve, valves.OnOffValve):
            if value not in (0, 1):
                raise ValueError("setpoint must be either 0 or 1 for an On/Off expiratory valve")
            elif value == 1:
                self._expiratory_valve.open()
            else:
                self._expiratory_valve.close()
        elif isinstance(self._expiratory_valve, valves.PWMControlValve):
            if not 0 <= value <= 100:
                raise ValueError(
                    "setpoint must be between 0 and 100 for an expiratory control valve"
                )
            else:
                self._expiratory_valve.setpoint = value
        self._setpoint_ex = value



##########################################################
###################
################### DoubleComp Lung Environment
###################
##########################################################

import torch
import numpy as np

from deluca.lung.core import LungEnv

# Sources:
# https://github.com/CohenLabPrinceton/Ventilator-Dev/blob/master/sandbox/HOWTO_RunController.ipynb
# https://github.com/CohenLabPrinceton/Ventilator-Dev/blob/master/vent/controller/control_module.py
# https://github.com/MinRegret/venti
def PropValve(x):  # copied from Controller.__SimulatedPropValve
    y = 3 * x
    y_torch = torch.from_numpy(np.array([0.03 * (y - 130)])).float()
    flow_new = 1.0 * (torch.tanh(y_torch) + 1)

    return torch.clamp(flow_new, 0.0, 1.72)


def Solenoid(x):  # copied from Controller.__SimulatedSolenoid
    if x > 0:
        return x / x
    else:
        return x * 0.0


class DoubleCompLung(LungEnv):
    def __init__(
        self, resistances=(0.1, 0.1), compliances=(0.1, 0.1), min_volume=5, peep_value=5, **kwargs
    ):
        # dynamics hyperparameters
        self.min_volume = min_volume
        self.P0 = peep_value
        self.peep_value = peep_value
        self.R1 = resistances[0]
        self.R2 = resistances[1]
        self.C1 = compliances[0]
        self.C2 = compliances[1]
        self.flow_pvs = 0.0
        self.flow = 0.0
        self.pressure_pvs = peep_value
        self.pressure = peep_value

        # reset states
        self.reset()

    def reset(self):
        self.volume = self.min_volume
        self.flow_pvs = 0.0
        self.flow = 0.0
        self.pressure_pvs = self.peep_value
        self.pressure = self.peep_value

    def compute_aux_states(self):
        # compute dp and df
        p_dot = (self.pressure - self.pressure_pvs) / self.dt
        f_dot = (self.flow - self.flow_pvs) / self.dt
        r1 = self.R1
        r2 = self.R2
        e1 = 1 / self.C1
        e2 = 1 / self.C2

        # compute pressure
        self.pressure = (
            (
                (f_dot * r1 * r2)
                + (self.flow * ((e2 * r1) + (e1 * (r1 + r2))))
                + (self.volume * e1 * e2)
                - (p_dot * r2)
            )
        ) / e2

    def step(self, u_in, u_out):
        dt = self.dt

        # calculate flow from u_in
        self.flow_pvs = self.flow
        self.flow = torch.clamp(PropValve(u_in), 0, 2)

        # update by flow rate
        self.volume = self.volume + self.flow * dt

        # compute and record state
        self.compute_aux_states()

        return self.pressure



from deluca.lung.core import LungEnv
from deluca.lung.core import Controller
from deluca.lung.core import ROOT
from deluca.lung.utils.core import BreathWaveform


__all__ = [
    "Controller",
    "LungEnv",
    "ROOT",
    "BreathWaveform",
]



import deluca.core
from deluca.lung.core import Controller, ControllerState
from deluca.lung.utils import BreathWaveform
from deluca.lung.controllers import Expiratory
from deluca.lung.environments._stitched_sim import StitchedSimObservation
from deluca.lung.utils.data.transform import ShiftScaleTransform
from deluca.lung.core import DEFAULT_DT
from deluca.lung.core import proper_time
import jax
import jax.numpy as jnp
import optax
import numpy as np
import itertools
import torch
import flax.linen as fnn


DEFAULT_DT = 0.03
class Deep_network(fnn.Module):
    H: int = 100
    kernel_size: int = 5
    out_dim: int = 1

    @fnn.compact
    def __call__(self, x):
        x = fnn.Conv(features=self.H, kernel_size=self.kernel_size, name=f"deep_conv")(x)
        x = fnn.relu(x)
        x = x.reshape((x.shape[0], -1))  # flatten
        x = fnn.Dense(features=1, use_bias=True, name=f"deep_fc")(x)
        return x

class DeepControllerState(deluca.Obj):
    waveform: deluca.Obj # waveform has to be here because it is subject to change during training
    errs: jnp.array
    time: float = float("inf")
    steps: int = 0
    dt: float = DEFAULT_DT
    
class Deep(Controller):
    params: list = deluca.field(jaxed=True)
    model: fnn.module = deluca.field(Deep_network, jaxed=False)
    featurizer: jnp.array = deluca.field(jaxed=False)
    H: int = deluca.field(100, jaxed=False)
    input_dim: int = deluca.field(1, jaxed=False)
    history_len: int = deluca.field(10, jaxed=False)
    kernel_size: int = deluca.field(5, jaxed=False)
    clip: float = deluca.field(40.0, jaxed=False)
    normalize: bool = deluca.field(False, jaxed=False)
    u_scaler: ShiftScaleTransform = deluca.field(jaxed=False)
    p_scaler: ShiftScaleTransform = deluca.field(jaxed=False)
    # bptt: int = deluca.field(1, jaxed=False) not used right now
    # TODO: add analogue of activation=torch.nn.ReLU

    def setup(self):
        self.model = Deep_network(H=self.H, kernel_size=self.kernel_size, out_dim=1)
        if self.params is None:
            self.params = self.model.init(
                jax.random.PRNGKey(0), jnp.expand_dims(jnp.ones([self.history_len]), axis=(0,1))
            )["params"]

        # linear feature transform:
        # errs -> [average of last h errs, ..., average of last 2 errs, last err]
        # emulates low-pass filter bank

        self.featurizer = jnp.tril(jnp.ones((self.history_len, self.history_len)))
        self.featurizer /= jnp.expand_dims(jnp.arange(self.history_len, 0, -1), axis = 0)

        if self.normalize:
            self.u_scaler = u_scaler
            self.p_scaler = p_scaler

    def init(self, waveform=BreathWaveform.create()):
        errs = jnp.array([0.0] * self.history_len)
        state = DeepControllerState(errs=errs, waveform=waveform)
        return state

    def __call__(self, controller_state, obs):
        state, t = obs.predicted_pressure, obs.time
        errs, waveform = controller_state.errs, controller_state.waveform
        target = waveform.at(t)
        
        if self.normalize:
            target_scaled = self.p_scaler(target).squeeze()
            state_scaled = self.p_scaler(state).squeeze()
            next_errs = jnp.roll(errs, shift=-1)
            next_errs = next_errs.at[-1].set(target_scaled - state_scaled)
        else:
            next_errs = jnp.roll(errs, shift=-1)
            next_errs = next_errs.at[-1].set(target - state)
        controller_state = controller_state.replace(errs=next_errs)

        decay = waveform.decay(t)

        def true_func(null_arg):
            trajectory = jnp.expand_dims(next_errs[-self.history_len:], axis=(0,1))
            u_in = self.model.apply({"params": self.params}, (trajectory @ self.featurizer))
            return u_in.squeeze().astype(jnp.float32)
        # changed decay compare from None to float(inf) due to cond requirements
        u_in = jax.lax.cond(jnp.isinf(decay), 
                            true_func,
                            lambda x : jnp.array(decay),
                            None)

        u_in = jax.lax.clamp(0.0, u_in.astype(jnp.float32), self.clip).squeeze()

        # update controller_state
        new_dt = jnp.max(jnp.array([DEFAULT_DT, t - proper_time(controller_state.time)]))
        new_time = t
        new_steps = controller_state.steps + 1
        controller_state = controller_state.replace(time=new_time, steps=new_steps, dt=new_dt)
        return controller_state, u_in
    '''
    def train_global(
        self,
        sims,
        pip_feed="parallel",
        duration=3,
        dt=0.03,
        epochs=100,
        use_noise=False,
        optimizer=torch.optim.Adam,
        optimizer_params={"lr": 1e-3, "weight_decay": 1e-4},
        loss_fn=torch.nn.L1Loss,
        loss_fn_params={},
        scheduler=torch.optim.lr_scheduler.ReduceLROnPlateau,
        scheduler_params={"factor": 0.9, "patience": 10},
        use_tqdm=True,
        print_loss=1,
        shuffle=False,
        device="cpu",
    ):
        optimizer = optimizer(self.parameters(), **optimizer_params)
        scheduler = scheduler(optimizer, **scheduler_params)
        loss_fn = loss_fn(**loss_fn_params)

        tt = torch.linspace(0, duration, int(duration / dt))
        losses = []

        torch.autograd.set_detect_anomaly(True)

        PIPs = [10, 15, 20, 25, 30, 35]
        PEEP = 5

        # TODO: handle device-awareness
        for epoch in range(epochs):

            if pip_feed == "parallel":
                self.zero_grad()
                loss = torch.tensor(0.0, device=device, requires_grad=True)

            for PIP, sim in itertools.product(PIPs, sims):

                if pip_feed == "sequential":
                    self.zero_grad()
                    loss = torch.tensor(0.0, device=device, requires_grad=True)

                self.waveform = BreathWaveform((PEEP, PIP))
                expiratory = Expiratory(waveform=self.waveform)

                self.reset()
                sim.reset()

                for t in tt:
                    sim.pressure += use_noise * torch.normal(mean=torch.tensor(1.5), std=1.0)
                    pressure = sim.pressure
                    u_in = self(pressure, self.waveform.at(t), t)
                    u_out = expiratory(pressure, self.waveform.at(t), t)
                    sim.step(
                        u_in, u_out
                    )  # potentially add multiplicative noise by * torch.normal(mean=torch.tensor(1.5), std=0.5)

                    if u_out == 0:
                        loss = loss + loss_fn(torch.tensor(self.waveform.at(t)), pressure)

                if pip_feed == "sequential":
                    loss.backward(retain_graph=True)
                    optimizer.step()
                    scheduler.step(loss)
                    per_step_loss = loss / len(tt)
                    losses.append(per_step_loss)

                    if epoch % print_loss == 0:
                        print(
                            f"Epoch: {epoch}, PIP: {PIP}\tLoss: {per_step_loss:.2f}\tLR: {optimizer.param_groups[0]['lr']}"
                        )

            if pip_feed == "parallel":
                loss.backward(retain_graph=True)
                optimizer.step()
                scheduler.step(loss)
                per_step_loss = loss / len(tt)
                losses.append(per_step_loss)

                if epoch % print_loss == 0:
                    print(
                        f"Epoch: {epoch}\tLoss: {per_step_loss:.2f}\tLR: {optimizer.param_groups[0]['lr']}"
                    )

        return losses
    '''

def rollout(controller, sim, tt, use_noise, PEEP, PIP, loss_fn, loss):
    waveform = BreathWaveform.create(custom_range=(PEEP, PIP))
    expiratory = Expiratory.create(waveform=waveform)
    controller_state = controller.init(waveform)
    expiratory_state = expiratory.init()
    sim_state, obs = sim.reset() 
    def loop_over_tt(ctrlState_expState_simState_obs_loss, t):
        controller_state, expiratory_state, sim_state, obs, loss = ctrlState_expState_simState_obs_loss
        mean = 1.5
        std = 1.0
        noise = mean + std * jax.random.normal(jax.random.PRNGKey(0), shape=())
        pressure = sim_state.predicted_pressure + use_noise * noise
        sim_state = sim_state.replace(predicted_pressure=pressure) # Need to update p_history as well or no?
        obs = obs.replace(predicted_pressure=pressure, time=t)

        controller_state, u_in = controller(controller_state, obs)
        expiratory_state, u_out = expiratory(expiratory_state, obs)

        sim_state, obs = sim(sim_state, (u_in, u_out))
        loss = jax.lax.cond(u_out == 0,
                            lambda x: x + loss_fn(jnp.array(waveform.at(t)), pressure),
                            lambda x: x,
                            loss)
        return (controller_state, expiratory_state, sim_state, obs, loss), None
    (_, _, _, _, loss), _ = jax.lax.scan(loop_over_tt, (controller_state, expiratory_state, sim_state, obs, loss), tt)
    return loss
    

def rollout_parallel(controller, sim, tt, use_noise, PEEP, PIPs, loss_fn):
    loss = jnp.array(0.)
    for PIP in PIPs:
        loss = rollout(controller, sim, tt, use_noise, PEEP, PIP, loss_fn, loss)
    return loss

# TODO: add scheduler and scheduler_params
# Question: Jax analogue of torch.autograd.set_detect_anomaly(True)?
def deep_train(
    controller,
    sim,
    pip_feed="parallel",
    duration=3,
    dt=0.03,
    epochs=100,
    use_noise=False,
    optimizer=optax.adamw,
    optimizer_params={"learning_rate": 1e-3, "weight_decay": 1e-4},
    loss_fn=lambda x, y: (jnp.abs(x - y)).mean(),
    # scheduler=torch.optim.lr_scheduler.ReduceLROnPlateau,
    # scheduler_params={"factor": 0.9, "patience": 10},
    print_loss=1,
):
    optim = optimizer(**optimizer_params)
    optim_state = optim.init(controller)

    tt = jnp.linspace(0, duration, int(duration / dt))
    losses = []

    # torch.autograd.set_detect_anomaly(True)

    PIPs = [10, 15, 20, 25, 30, 35]
    PEEP = 5

    # TODO: handle device-awareness
    for epoch in range(epochs):

        if pip_feed == "parallel":
            value, grad = jax.value_and_grad(rollout_parallel)(controller, sim, tt, use_noise, PEEP, PIPs, loss_fn)
            updates, optim_state = optim.update(grad, optim_state, controller)
            controller = optax.apply_updates(controller, updates)
            per_step_loss = value / len(tt)
            losses.append(per_step_loss)

            if epoch % print_loss == 0:
                print(
                    f"Epoch: {epoch}\tLoss: {per_step_loss:.2f}"
                )

        if pip_feed == "sequential":
            for PIP in PIPs:
                value, grad = jax.value_and_grad(rollout)(controller, sim, tt, use_noise, PEEP, PIP, loss_fn, jnp.array(0.))
                updates, optim_state = optim.update(grad, optim_state, controller)
                controller = optax.apply_updates(controller, updates)
                per_step_loss = value / len(tt)
                losses.append(per_step_loss)

                if epoch % print_loss == 0:
                    print(
                        f"Epoch: {epoch}, PIP: {PIP}\tLoss: {per_step_loss:.2f}"
                    )
    return controller




import numpy as np
import tqdm
import datetime

from deluca.lung.controllers import PID
from deluca.lung.environments import PhysicalLung
from deluca.lung.utils import Analyzer, BreathWaveform

from deluca.lung.utils import run_calibration
from deluca.lung.utils import run_controller

DEFAULT_VALUES = np.concatenate((np.linspace(0, 0.9, 10), np.linspace(1, 10, 10)))


def run_pid_grid(
    Ps=DEFAULT_VALUES,
    Is=DEFAULT_VALUES,
    Ds=[0.0],
    PIPs=[10, 15, 20, 25, 30, 35],
    R=50,
    C=10,
    PEEP=5,
    abort=60,
    T=300,
    directory=None,
    env=None,
    **kwargs,
):
    analyzers = []
    env = env or PhysicalLung()

    print("Running calibration")
    run_calibration(R, C, PEEP, directory, env=env)

    print("Running grid")
    for PIP in tqdm.tqdm(PIPs):
        for P in tqdm.tqdm(Ps, position=0, leave=False):
            for I in Is:
                for D in Ds:
                    env.reset()

                    waveform = BreathWaveform((PEEP, PIP))
                    pid = PID(K=[P, I, D], waveform=waveform)
                    result = run_controller(
                        pid, R, C, T, abort, env=env, waveform=waveform, directory=directory
                    )
                    analyzers.append(Analyzer(result))

    print("Running calibration")
    run_calibration(R, C, PEEP, directory, env=env)

    return analyzers



# NOTE: API needs work (e.g., what should __call__'s signature be)

import torch

from deluca.lung.utils.core import TorchStandardScaler


class Featurizer:
    def reset(self):
        raise NotImplementedError()

    def __call__(self, *args, **kwargs):
        raise NotImplementedError()


# Simulator featurizers


class ScalingHistoryFeaturizer(Featurizer):
    def __init__(self, u_window, p_window, u_scaler, p_scaler):
        self.u_window = u_window
        self.p_window = p_window
        self.u_scaler = TorchStandardScaler(u_scaler)
        self.p_scaler = TorchStandardScaler(p_scaler)

        self.reset()

    def reset(self):
        self.u_history = [self.u_scaler.transform(0).detach()] * self.u_window
        self.p_history = [self.p_scaler.transform(0).detach()] * self.p_window

    def __call__(self, u_in, u_out, pressure, *args, **kwargs):
        if not isinstance(u_in, torch.Tensor):
            u_in = torch.tensor(u_in)
        u_in = self.u_scaler.transform(u_in)
        self.u_history.append(u_in)
        self.u_history.pop(0)

        if not isinstance(pressure, torch.Tensor):
            pressure = torch.tensor(pressure)
        pressure = self.p_scaler.transform(pressure)
        self.p_history.append(pressure)
        self.p_history.pop(0)

        return torch.cat([torch.cat(self.u_history), torch.cat(self.p_history)])

# Controller featurizers


class TriangleErrorFeaturizer(Featurizer):
    def __init__(self, history_len):
        self.history_len = history_len

        self.coef = torch.ones((history_len, history_len), requires_grad=False).tril_()
        self.coef /= torch.arange(history_len, 0, -1).unsqueeze(0)

        self.reset()

    def reset(self):
        self.errs = [torch.tensor(0., dtype=torch.get_default_dtype())] * self.history_len

    def __call__(self, pressure, target, t, *args, **kwargs):
        self.errs.append(target - pressure)
        self.errs.pop(0)

        trajectory = torch.stack(self.errs).unsqueeze(0).unsqueeze(0)

        return trajectory @ self.coef



import flax.linen as fnn
from deluca.lung.utils.data.munger import Munger
from deluca.lung.utils.data.alpha_dropout import Alpha_Dropout

# TODO: learning rate scheduler for Stitched_sim_trainer
class SNN(fnn.Module):
    out_dim: int = 1
    hidden_dim: int = 100
    n_layers: int = 4
    dropout_prob: float = 0.0
    scale: float = 1.0507009873554804934193349852946
    alpha: float = 1.6732632423543772848170429916717

    @fnn.compact
    def __call__(self, x):
        for i in range(self.n_layers - 1):
            x = fnn.Dense(features=self.hidden_dim, use_bias=False, name=f"SNN_fc{i}")(x)
            x = self.scale * fnn.elu(x, alpha=self.alpha)
            x = Alpha_Dropout(rate=self.dropout_prob, deterministic=False)(x)
        x = fnn.Dense(features=self.out_dim, use_bias=True, name=f"SNN_fc{i + 1}")(x)
        return x


class ShallowBoundaryModel(fnn.Module):
    out_dim: int = 1
    hidden_dim: int = 100
    model_num: int = 0

    @fnn.compact
    def __call__(self, x):
        x = fnn.Dense(
            features=self.hidden_dim,
            use_bias=False,
            name=f"shallow_fc{1}_model" + str(self.model_num),
        )(x)
        x = fnn.tanh(x)
        x = fnn.Dense(
            features=self.out_dim, use_bias=True, name=f"shallow_fc{2}_model" + str(self.model_num)
        )(x)
        return x


class ConstantModel(fnn.Module):
    const_out: float = 0.0

    @fnn.compact
    def __call__(self, x):
        return self.const_out

    def update_constant(self, const_out):
        self.const_out = const_out



from deluca.lung.experimental.controllers._ibonus import IBonus, IBonusCubic, LookaheadPID
# from deluca.lung.experimental.controllers._deep_pid_residual import DeepPIDResidual
# from deluca.lung.experimental.controllers._deep_pid_residual_clipped import DeepPIDResidualClipped
from deluca.lung.experimental.controllers._pid_correction import PIDCorrection
from deluca.lung.experimental.controllers._deep import Deep, deep_train


import torch

from deluca.lung.core import TorchLungEnv


class Linear(TorchLungEnv):
    def __init__(self, featurizer):
        self.featurizer = featurizer
        self.network = torch.nn.Linear(featurizer.u_window + featurizer.p_window, 1, bias=True)

        self.reset()

    def reset(self):
        self.featurizer.reset()
        self.pressure = 0

    def step(self, u_in, u_out):
        """
        For inference and returns unscaled pressure
        """
        features = self.featurizer(u_in, u_out, self.pressure)
        self.pressure = self.featurizer.p_scaler.inverse_transform(self.network(features)).squeeze()
        return self.pressure

    def forward(self, features):
        """
        For training and returns scaled pressure
        """
        return self.network(features)



from deluca.lung.utils.core import BreathWaveform
from deluca.lung.utils.data.analyzer import Analyzer
from deluca.lung.utils.data.munger import Munger
# from deluca.lung.utils.data.featurizer import Featurizer
# from deluca.lung.utils.data.featurizer import ScalingHistoryFeaturizer
# from deluca.lung.utils.data.featurizer import TriangleErrorFeaturizer
from deluca.lung.utils.sim.nn import SNN
from deluca.lung.utils.sim.nn import ShallowBoundaryModel
from deluca.lung.utils.sim.nn import ConstantModel
from deluca.lung.utils.sim.nn_jax import InspiratoryModel_jax
from deluca.lung.utils.sim.testing import open_loop_test
# from deluca.lung.utils.scripts.run_calibration import run_calibration
from deluca.lung.utils.scripts.save_data_and_plot import save_data_and_plot
from deluca.lung.utils.scripts.run_controller import run_controller, run_controller_scan
from deluca.lung.utils.scripts.find_best_pid import find_best_pid, find_global_best_pid, plot_pid
# from deluca.lung.utils.scripts.run_explorer import run_explorer
# from deluca.lung.utils.scripts.run_pid_grid import run_pid_grid
from deluca.lung.utils.scripts.train_controller import train_controller, train_controller_multipip
from deluca.lung.utils.scripts.train_simulator import train_simulator
from deluca.lung.utils.scripts.convert_venti import convert_sim

__all__ = [
    "BreathWaveform",
    "Analyzer",
    "Munger",
    # "Featurizer",
    # "ScalingHistoryFeaturizer",
    # "TriangleErrorFeaturizer",
    "SNN",
    "ShallowBoundaryModel",
    "ConstantModel",
    "InspiratoryModel_jax",
    "open_loop_test",
    # "run_calibration",
    "save_data_and_plot",
    "run_controller",
    "find_best_pid",
    "find_global_best_pid",
    "plot_pid",
    # "run_explorer",
    # "run_pid_grid",
    "train_controller",
    "train_simulator",
    "convert_sim",
    "train_controller_multipip",
]



from deluca.lung.devices import IODeviceBase
from deluca.lung.devices.base import gpio_command


class Pin(IODeviceBase):
    """ Base Class wrapping pigpio methods for interacting with GPIO pins on
    the raspberry pi. Subclasses include InputPin, OutputPin; along with
    any specialized pins or specific devices defined in vent.io.actuators
    & vent.io.sensors (note: actuators and sensors do not need to be tied
    to a GPIO pin and may instead be interfaced through an ADC or I2C).

    This is an abstract base class. The subclasses InputPin and
    OutputPin extend Pin into a usable form.
    """
    _PIGPIO_MODES = {'INPUT': 0,
                     'OUTPUT': 1,
                     'ALT5': 2,
                     'ALT4': 3,
                     'ALT0': 4,
                     'ALT1': 5,
                     'ALT2': 6,
                     'ALT3': 7}

    def __init__(self, pin, gpio=None):
        """ Inherits attributes and methods from IODeviceBase.

        Args:
            pin (int): The number of the pin to use
            gpio (PigpioConnection): pigpiod connection to use; if not specified, a new one is established
        """
        super().__init__(gpio)
        self.pin = pin

    @property
    @gpio_command
    def mode(self) -> str:
        """ The currently active pigpio mode of the pin."""
        return {value: key for key, value in self._PIGPIO_MODES.items()}[self.gpio.get_mode(self.pin)]

    @mode.setter
    @gpio_command
    def mode(self, mode):
        """ Performs validation on requested mode, then sets the mode. Raises runtime error if something goes wrong.

        Args:
            mode (str): A mode in _PIGPIO_MODES
        """
        if mode not in self._PIGPIO_MODES:
            raise ValueError("Pin mode must be one of: {}".format([*self._PIGPIO_MODES.keys()]))
        result = self.gpio.set_mode(self.pin, self._PIGPIO_MODES[mode])
        if result != 0:
            raise RuntimeError('Failed to write mode {} on pin {}'.format(mode, self.pin))

    def toggle(self):
        """ If pin is on, turn it off. If it's off, turn it on. Do not raise a warning when pin is read in this way."""
        self.write(not self.read())

    @gpio_command
    def read(self) -> int:
        """ Returns the value of the pin: usually 0 or 1 but can be overridden by subclass."""
        return self.gpio.read(self.pin)

    @gpio_command
    def write(self, value):
        """ Sets the value of the Pin. Usually 0 or 1 but behavior differs for some subclasses.

        Args:
            value: The value to write to the pin. Can be either `1` to turn on the pin or `0` to turn it off.
        """
        if value not in (0, 1):
            raise ValueError('Cannot write a value other than 0 or 1 to a Pin')
        self.gpio.write(self.pin, value)


class PWMOutput(Pin):
    """ A pin configured to output a PWM signal. Can be configured to use either a hardware-generated or
    software-generated signal. Overrides parent methods read() and write().
    """
    _DEFAULT_FREQUENCY = 20000
    _DEFAULT_SOFT_FREQ = 2000
    _HARDWARE_PWM_PINS = (12, 13, 18, 19)

    def __init__(self, pin, initial_duty=0, frequency=None, gpio=None):
        """ Inherits attributes from parent Pin, then sets PWM frequency & initial duty (use defaults if None)

        Args:
            pin (int): The number of the pin to use. Hardware PWM pins are 12, 13, 18, and 19.
            initial_duty (float): The initial duty cycle of the pin. Must be between 0 and 1.
            frequency (float): The PWM frequency to use.
            pig (PigpioConnection): pigpiod connection to use; if not specified, a new one is established
        """
        super().__init__(pin=pin, gpio=gpio)
        self._hardware_enabled = True if self.pin in self._HARDWARE_PWM_PINS else False
        default_f = self._DEFAULT_FREQUENCY if self._hardware_enabled else self._DEFAULT_SOFT_FREQ
        self.__pwm(default_f if frequency is None else frequency, initial_duty)

    @property
    def hardware_enabled(self):
        """ Return true if this is a hardware-enabled PWM pin; False if not. The Raspberry Pi only supports hardware-
        generated PWM on pins 12, 13, 18, and 19, so generally `hardware_enabled` will be true if this is one of those,
        and false if it is not. However, `hardware_enabled` can also by dynamically set to False if for some reason
        pigpio is unable to start a hardware PWM (i.e. if the clock is unavailable or in use or something)
        """
        return self._hardware_enabled

    @property
    @gpio_command
    def frequency(self) -> float:
        """ Return the current PWM frequency active on the pin."""
        return self.gpio.get_PWM_frequency(self.pin)

    @frequency.setter
    def frequency(self, new_frequency):
        """ TODO: extend exception handling/logging to this
        Note: pigpio.pi.hardware_PWM() returns 0 if OK and an error code otherwise.
        - Tries to write hardware PWM if hardware_enabled
        - If that fails, or if not hardware_enabled, tries to write software PWM instead.

        Args:
            new_frequency (float): A new PWM frequency to use.
            """
        self.__pwm(new_frequency, self._duty())

    @property
    @gpio_command
    def duty(self) -> float:
        """ Returns the PWM duty cycle (pulled straight from pigpiod) mapped to the range [0-1] """
        return self.gpio.get_PWM_dutycycle(self.pin) / self.gpio.get_PWM_range(self.pin)

    @gpio_command
    def _duty(self) -> int:
        """ Returns the pigpio integer representation of the duty cycle."""
        return self.gpio.get_PWM_dutycycle(self.pin)

    @duty.setter
    @gpio_command
    def duty(self, duty_cycle):
        """ Sets the duty cycle.
        Args:
            duty_cycle (float): The PWM duty cycle to set. Must be between 0 and 1 (verified upon calling __pwm()).
        """
        self.__pwm(self.frequency, duty_cycle)

    def read(self) -> float:
        """ Overridden to return duty cycle instead of reading the value on the pin."""
        return self.duty

    def write(self, value):
        """ Overridden to write duty cycle.

        Args:
            value (float): See `PWMOutput.duty`
        """
        self.duty = value

    def __pwm(self, frequency, duty):
        """ Sets a PWM frequency and duty using either hardware or software generated PWM according to the value of
        `self.hardware_enabled`. If hardware_enabled, starts a hardware pwm with the requested duty. If not
        hardware_enabled, or if there is a problem setting a hardware generated PWM, starts a software PWM.

        Args:
            frequency (float): A new PWM frequency to use.
            duty (float): The PWM duty cycle to set. Must be between 0 and 1.
        """
        if not 0 <= duty <= 1:
            raise ValueError('Duty cycle must be between 0 and 1 but got {}'.format(duty))
        _duty = int(duty * self.gpio.get_PWM_range(self.pin))
        if self._hardware_enabled:
            self.__hardware_pwm(frequency, _duty)
        if not self._hardware_enabled:
            self.__software_pwm(frequency, _duty)

    @gpio_command
    def __hardware_pwm(self, frequency, duty):
        """ Used for pins where hardware pwm is available.
        -Tries to write a hardware pwm. result == 0 if it succeeds.
        -Sets hardware_enabled flag to indicate success or failure

        Args:
            frequency: A new PWM frequency to use.
            duty (int): The PWM duty cycle to set. Must be between 0 and 1.
        """
        try:
            self.gpio.hardware_PWM(self.pin, frequency, duty)
            self._hardware_enabled = True
        except Exception as e:
            self._hardware_enabled = False
            self.__software_pwm(frequency, duty)

    @gpio_command
    def __software_pwm(self, frequency, duty):
        """ Used for pins where hardware PWM is NOT available.

        Args:
            frequency: A new PWM frequency to use.
            duty (int): A pigpio integer representation of duty cycle
        """
        self.gpio.set_PWM_dutycycle(self.pin, duty)
        realized_frequency = self.gpio.set_PWM_frequency(self.pin, frequency)
        if frequency != realized_frequency:
            raise RuntimeWarning(
                'A PWM frequency of {} was requested but the best that could be done was {}'.format(
                    frequency,
                    realized_frequency
                )
            )
        self._hardware_enabled = False



import jax.numpy as jnp
import numpy as np
import sklearn.preprocessing
import matplotlib.pyplot as plt
import pandas as pd

from deluca.lung.utils.data.analyzer import Analyzer
from deluca.lung.utils.data.jax_data import *
from deluca.lung.utils.data.transform import *
import tensorflow as tf

# NOTES
# - @classmethod factory


def resample(tt, f, dt=0.03, snap_left=False, return_grid=False):
    # resample a continuous-time function interpolated from non-uniform samples
    num_tt = int((tt[-1] - tt[0]) / dt)
    tt_grid = jnp.arange(num_tt) * dt + tt[0]

    if snap_left:
        f_grid = f[jnp.searchsorted(tt, tt_grid)]
    else:
        f_grid = jnp.interp(tt_grid, tt, f)

    if return_grid:
        return tt_grid, f_grid
    else:
        return f_grid


class Munger:
    @classmethod
    def from_kaggle(cls, path, oiasdjfoaisdjf):
        obj = cls(path)

        return obj

    @classmethod
    def from_legacy(cls, paths, oiasdjfoij):
        obj = cls(paths)

        return obj

    def __init__(
        self,
        paths=None,
        kagglePath_or_df=None,
        R=None,
        C=None,
        to_round=False,
        dt_resample=False,
        clip=(2, -1),
        inspiratory_only=True,
        splits=[0.9, 0.1],
        fit_scaler_key="train",
        truncate=True,
        **kwargs,
    ):
        self.paths = paths
        self.analyzers = []
        self.data = []
        self.splits = {}
        self.to_round = to_round

        # Use Kaggle Dataset
        if kagglePath_or_df is not None:
            self.add_data_jax(kagglePath_or_df, dt_resample, R=R, C=C, **kwargs)
        # Load dataset using legacy paths
        elif paths is not None:
            self.add_data(
                paths,
                dt_resample,
                clip,
                inspiratory_only=inspiratory_only,
                truncate=truncate,
                **kwargs,
            )
        # Shuffle and split data
        self.split_data(splits=splits)
        # Compute mean, std for u_in and pressure
        self.u_scaler, self.p_scaler = self.fit_scalers(key=fit_scaler_key)  # new_training

    # TODO: do we need to skip first 2 and last breath? Right now just reading from kaggle dataset
    def add_data_jax(self, kagglePath_or_df, dt_resample=True, R=None, C=None, **kwargs):
        failed = 0
        if isinstance(kagglePath_or_df, str):
            df = pd.read_csv(kagglePath_or_df)
            if not (R is None and C is None):
                df = df[(df["R"] == R) & (df["C"] == C)]
        else:
            df = kagglePath_or_df
        breath_ids = df["breath_id"].unique()
        num_breaths = len(breath_ids)
        # print('num_breaths:' + str(num_breaths))
        for breath_id in breath_ids:
            # print('--------------------------')
            # print('breath_id: ' + str(breath_id))
            try:
                df_breath_id = df[df["breath_id"] == breath_id]
                # print(df_breath_id.head())
                if self.to_round:
                    u_in = jnp.round(df_breath_id["u_in"].to_numpy())
                else:
                    u_in = df_breath_id["u_in"].to_numpy()

                pressure = df_breath_id["pressure"]

                if dt_resample:
                    tt = df_breath_id["time_step"]
                    u_in = resample(tt, u_in)
                    pressure = resample(tt, pressure)

                self.data.append((u_in, pressure))
                # print('u_in:' + str(u_in))
                # print('pressure' + str(pressure))

            except Exception as e:
                print(breath_id, e)
                failed += 1

        print(f"Added {len(self.data)} breaths from {num_breaths - failed} breath ids.")

    # Note: the following methods are meant to be run in order
    def add_data(
        self, paths, dt_resample=True, clip=(2, -1), inspiratory_only=True, truncate=True, **kwargs
    ):
        if isinstance(paths, str):
            paths = [paths]

        failed = 0
        kaggle_truncate_threshold = 29
        for path in paths:
            try:
                analyzer = Analyzer(path)
                self.analyzers.append(analyzer)
                if inspiratory_only:
                    clips = analyzer.infer_inspiratory_phases()
                else:
                    clips = analyzer.infer_breaths()

                for start, end in clips[clip[0] : clip[1]]:  # skip first 2 & last breaths
                    if self.to_round:
                        u_in = jnp.round(analyzer.u_in[start:end])
                    else:
                        u_in = analyzer.u_in[start:end]
                    pressure = analyzer.pressure[start:end]
                    if dt_resample:
                        tt = analyzer.tt[start:end]
                        u_in = resample(tt, u_in)
                        pressure = resample(tt, pressure)
                    if truncate:
                        if len(u_in) < kaggle_truncate_threshold:
                            continue
                        u_in = u_in[:kaggle_truncate_threshold]
                        pressure = pressure[:kaggle_truncate_threshold]
                    self.data.append((u_in, pressure))

            except Exception as e:
                print(path, e)
                failed += 1

        print(f"Added {len(self.data)} breaths from {len(self.paths) - failed} paths.")

    def split_data(self, seed=0, keys=["train", "test"], splits=[0.9, 0.1], **kwargs):
        rng = np.random.default_rng(seed)

        # Determine split boundaries
        splits = (np.array(splits) / np.sum(splits) * len(self.data)).astype("int")[:-1]

        # Everyday I'm shuffling
        rng.shuffle(self.data)

        # Splitting
        self.splits = {key: val for key, val in zip(keys, np.split(self.data, splits))}

    def fit_scalers(self, key="train"):
        u_scaler = ShiftScaleTransform(jnp.array([u for u, p in self.splits[key]]))
        print(f"u_in: mean={u_scaler.mean}, std={u_scaler.std}")
        p_scaler = ShiftScaleTransform(jnp.array([p for u, p in self.splits[key]]))
        print(f"pressure: mean={p_scaler.mean}, std={p_scaler.std}")

        return u_scaler, p_scaler

    def get_scaled_data(self, key, scale_data=False):
        X, y = [], []
        if scale_data:
            X = jnp.array([self.u_scaler(u) for u, p in self.splits[key]])
            y = jnp.array([self.p_scaler(p) for u, p in self.splits[key]])
        else:
            X = jnp.array([u for u, p in self.splits[key]])
            y = jnp.array([p for u, p in self.splits[key]])
        return X, y

    def get_dataloader_tf(self, key, shuffle=True, scale_data=False, batch_size=20):
        X, y = self.get_scaled_data(key, scale_data)
        # dataset = Dataset_from_XY(X, y)
        # return Dataloader_jax(dataset, batch_size=batch_size, shuffle=shuffle)
        dataset = tf.data.Dataset.from_tensor_slices((X, y))
        if shuffle:
            dataset = dataset.shuffle(buffer_size=X.shape[0])
        return dataset

    def unscale_pressures(self, p):  # TODO: make the jax version of this
        if type(p) is float:
            p = jnp.array([p])
        return self.p_scaler.inverse(p)[:, 0]

    ###########################################################################
    # Plotting methods
    ###########################################################################

    def plot_boundary_pressures(self):
        plt.rc("figure", figsize=(16, 4))

        for tau in range(1, 6):
            plt.subplot(150 + tau)

            u_init = []
            p_init = []

            for u, p in self.splits["train"]:
                u_init.append(u[:tau].mean())
                p_init.append(p[tau])

            plt.xlim([0, 105])
            plt.ylim([2, 33])
            plt.xlabel(f"u_in[0:{tau}].mean")
            plt.title(f"pressure[{tau}]")

            plt.scatter(u_init, p_init, s=1)

    def scale_and_window_boundary_jax(self, key, boundary_index):
        X, y = [], []

        if boundary_index == 0:  # special case: no features, predict p[0]
            for u, p in self.splits[key]:
                p_scaled = self.p_scaler(p[0])
                target = p_scaled
                y.append(target)
            return None, jnp.array(y)

        # otherwise, collate [first B inputs, first B pressures] -> (next pressure) pairs
        for u, p in self.splits[key]:
            T = len(u)
            if T < boundary_index + 1:  # if trajectory is too short, abort
                continue

            u_scaled = self.u_scaler(u[:boundary_index].reshape(-1, 1)).flat
            p_scaled = self.p_scaler(p[: boundary_index + 1].reshape(-1, 1)).flat

            features = jnp.concatenate([u_scaled, p_scaled[:-1]])
            target = p_scaled[-1]

            X.append(features)
            y.append(target)

        # return torch.tensor(X, dtype=self.dtype), torch.tensor(y, dtype=self.dtype)
        return jnp.array(X), jnp.array(y)

    def get_scaled_data_jax(self, key, scale_data=False):
        X, y = [], []
        if scale_data:
            X = jnp.array([self.u_scaler(u) for u, p in self.splits[key]])
            y = jnp.array([self.p_scaler(p) for u, p in self.splits[key]])
        else:
            X = jnp.array([u for u, p in self.splits[key]])
            y = jnp.array([p for u, p in self.splits[key]])
        return X, y
